# coverage: ignore
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Functions to determine optimal values for qro(a)m.

These were adapted from corresponding Mathematica functions which are given in
the docstrings.
"""
import numpy as np


def QR3(L, M1):
    r"""Find optimal cost for QROM based on L value of size M1.

    QR[Ll_, m_] := Ceiling[MinValue[{Ll/2^k + m*(2^k - 1), k >= 0}, k
    \[Element] Integers]];
    """
    k = 0.5 * np.log2(L / M1)
    value = lambda k: L / np.power(2, k) + M1 * (np.power(2, k) - 1)
    if k < 0:
        k_opt = 0
        val_opt = np.ceil(value(k_opt))
        assert val_opt.is_integer()
        return int(k_opt), int(val_opt)
    k_int = [np.floor(k), np.ceil(k)]  # restrict optimal k to integers
    k_opt = k_int[np.argmin(value(k_int))]  # obtain optimal k
    val_opt = np.ceil(value(k_opt))  # obtain ceiling of optimal value given k
    assert k_opt.is_integer()
    assert val_opt.is_integer()
    return int(k_opt), int(val_opt)


def QR2(L1, L2, M):
    """Find optimal cost for QROM over two values of L of size M.

     Table[Ceiling[L1/2^k1]*Ceiling[L2/2^k2] + M*(2^(k1 + k2) - 1), {k1, 1,
    10}, {k2, 1, 10}]
    """
    min_val = np.inf
    for k1 in range(1, 11):
        for k2 in range(1, 11):
            test_val = np.ceil(L1 /
                               (2**k1)) * np.ceil(L2 /
                                                  (2**k2)) + M * (2**
                                                                  (k1 + k2) - 1)
            if test_val < min_val:
                min_val = test_val
    return int(min_val)


def QI2(L1: int, Lv2: int):
    """Find optimal value for inverse QROM over two L values.

    QI2[L1_, L2_] :=
    Min[Table[
    Ceiling[L1/2^k1]*Ceiling[L2/2^k2] + 2^(k1 + k2), {k1, 1, 10}, {k2,
      1, 10}]];
    """
    min_val = np.inf
    for k1 in range(1, 11):
        for k2 in range(1, 11):
            test_val = np.ceil(L1 / (2**k1)) * np.ceil(Lv2 /
                                                       (2**k2)) + 2**(k1 + k2)
            if test_val < min_val:
                min_val = test_val
    return int(min_val)



# coverage: ignore
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from functools import reduce

import numpy as np
import pytest

from openfermion.resource_estimates import HAVE_DEPS_FOR_RESOURCE_ESTIMATES
if HAVE_DEPS_FOR_RESOURCE_ESTIMATES:
    from pyscf.pbc import gto, mp, scf
    from openfermion.resource_estimates.pbc.hamiltonian import (
        cholesky_from_df_ints,)
    from openfermion.resource_estimates.pbc.thc.factorizations.thc_jax import (
        kpoint_thc_via_isdf,)
    from openfermion.resource_estimates.pbc.thc.compute_lambda_thc import (
        compute_lambda,)
    from openfermion.resource_estimates.pbc.thc.thc_integrals import (
        KPTHCDoubleTranslation,)


@pytest.mark.skipif(not HAVE_DEPS_FOR_RESOURCE_ESTIMATES,
                    reason='pyscf and/or jax not installed.')
@pytest.mark.slow
def test_kpoint_thc_lambda():
    cell = gto.Cell()
    cell.atom = """
    C 0.000000000000   0.000000000000   0.000000000000
    C 1.685068664391   1.685068664391   1.685068664391
    """
    cell.basis = "gth-szv"
    cell.pseudo = "gth-hf-rev"
    cell.a = """
    0.000000000, 3.370137329, 3.370137329
    3.370137329, 0.000000000, 3.370137329
    3.370137329, 3.370137329, 0.000000000"""
    cell.unit = "B"
    cell.verbose = 0
    cell.mesh = [11] * 3
    cell.build()

    kmesh = [1, 1, 2]
    kpts = cell.make_kpts(kmesh)
    mf = scf.KRHF(cell, kpts)
    mf.kernel()
    #
    # Build kpoint THC eris
    #
    #
    rsmf = scf.KRHF(mf.cell, mf.kpts).rs_density_fit()
    # Force same MOs as FFTDF at least
    rsmf.mo_occ = mf.mo_occ
    rsmf.mo_coeff = mf.mo_coeff
    rsmf.mo_energy = mf.mo_energy
    rsmf.with_df.mesh = mf.cell.mesh
    mymp = mp.KMP2(rsmf)
    Luv = cholesky_from_df_ints(mymp)
    cthc = 4
    num_thc = cthc * mf.mo_coeff[0].shape[-1]
    np.random.seed(7)
    kpt_thc, _ = kpoint_thc_via_isdf(
        mf,
        Luv,
        num_thc,
        perform_adagrad_opt=False,
        perform_bfgs_opt=True,
        bfgs_maxiter=10,
        verbose=False,
    )
    hcore_ao = mf.get_hcore()
    hcore_mo = np.asarray([
        reduce(np.dot, (mo.T.conj(), hcore_ao[k], mo))
        for k, mo in enumerate(mf.mo_coeff)
    ])
    helper = KPTHCDoubleTranslation(kpt_thc.chi, kpt_thc.zeta, mf)
    lambda_data = compute_lambda(
        hcore_mo,
        helper,
    )
    assert np.isclose(lambda_data.lambda_total, 93.84613761765415)



#coverage:ignore
"""Compute qubit vs toffoli for THC LCU"""
from math import pi
import itertools
import numpy as np
import matplotlib.pyplot as plt
from numpy.lib.scimath import arccos, arcsin  # has analytc continuation to cplx
from openfermion.resource_estimates.utils import QR, QI


def qubit_vs_toffoli(lam,
                     dE,
                     eps,
                     n,
                     chi,
                     beta,
                     M,
                     algorithm='half',
                     verbose=False):
    """
    Args:
        lam (float) - the lambda-value for the Hamiltonian
        dE (float) - allowable error in phase estimation. usually 0.001
        eps (float) - allowable error for synthesis (dE/(10 * lam)) usually
        n (int) - number of spin orbitals.
        chi (int) - number of bits of precision for state prep
        beta (int) - number of bits of precision for rotations
        M (int) - THC rank or r_{Thc}
        algorithm (str) - 'half', where half of the phasing angles are loaded
                                  at a time
                          'full', where angles loaded from QROM to perform
                                  phasing operations are loaded at the same time
                          Note: In arXiv:2011.03494,
                              'half' corresponds to Fig 11, while
                              'full' corresponds to Fig 12.
        verbose (bool) - do additional printing of intermediates?

    """
    # only valid algorithms accepted
    assert algorithm in ['half', 'full']

    # The number of iterations for the phase estimation.
    iters = np.ceil(pi * lam / (dE * 2))
    # The number of bits used for each register.
    nM = np.ceil(np.log2(M + 1))
    #This is number of distinct items of data we need to output, see Eq. (28).
    d = M * (M + 1) / 2 + n / 2
    # The number of bits used for the contiguous register.
    nc = np.ceil(np.log2(d))
    # The output size is 2*Log[M] for the alt values, χ for the keep value,
    # and 2 for the two sign bits.
    m = 2 * nM + 2 + chi

    # The next block of code finds the optimal number of bits to use for the
    # rotation angle for the amplitude amplification taking into account the
    # probability of failure and the cost of the rotations.
    oh = np.zeros(20, dtype=float)
    for p in range(1, 20 + 1):
        cos_term = arccos(np.power(2, nM) / np.sqrt(d) / 2)
        # print(cos_term)
        v = np.round(np.power(2, p) / (2 * pi) * cos_term)
        asin_term = arcsin(
            np.cos(v * 2 * pi / np.power(2, p)) * np.sqrt(d) / np.power(2, nM))
        sin_term = np.sin(3 * asin_term)**2
        oh[p - 1] = (20_000 * (1 / sin_term - 1) + 4 * p).real
    # br is the number of bits used in the rotation.
    br = np.argmin(oh) + 1
    # Next are the costs for the state preparation.
    cp1 = 2 * (10 * nM + 2 * br - 9)
    # There is cost 10*Log[M] for preparing the equal superposition over the
    # input registers. This is the costing from above Eq. (29).
    # This is the cost of computing the contiguous register and inverting it.
    # This is with a sophisticated scheme adding together triplets of bits.
    # This is the cost of step 2 in the list on page 14.
    cp2 = 2 * (nM**2 + nM - 1)
    # This is the cost of the QROM for the state preparation and its inverse.
    cp3 = QR(d, m)[1] + QI(d)[1]
    # The cost for the inequality test.
    cp4 = 2 * chi
    # The cost 2*nM for the controlled swaps.
    cp5 = 4 * nM
    # Then there is a cost of nM+1 for swapping the μ and ν registers, where
    # the +3 is because we need to control on two registers, and
    # control swap of the spin registers.
    cp6 = 2 * nM + 3
    #  The total cost in Eq. (33).
    CPCP = cp1 + cp2 + cp3 + cp4 + cp5 + cp6

    # Next are the costs for the select operation.
    # This is the cost of swapping based on the spin register. These costs are
    # from the list on page 15, and this is steps 1 and 7.
    cs1 = 2 * n
    k1 = 2**QI(M + n / 2)[0]
    cs2a = M + n / 2 - 2 + np.ceil(M / k1) + np.ceil(n / 2 / k1) + k1

    # The QROM for the rotation angles the first time.  Here M+n/2-2 is the cost
    # for generating them, and the second part is the cost for inverting them
    # with advanced QROM. The QROM for the rotation angles the second time.
    # Here the cost M-2 is for generating the angles the second time, and QI[M]
    # is for inverting the QROM. Steps 2 and 6.
    cs2b = M - 2 + QI(M)[1]
    # The cost of the rotations steps 3 and 5.
    cs3 = 4 * n * (beta - 2)
    # Cost for extra part in making the Z doubly controlled step 4.
    cs4 = 1
    # The total select cost in Eq. (43).
    CS = cs1 + cs2a + cs2b + cs3 + cs4
    # The cost given slightly above Eq. (44) is 2*nM+5. That is a typo and it
    # should have the aleph (equivalent to χ here) like at the top of the
    # column. Here we have +3 in this line, +1 in the next line and +1 for cs4,
    #  to give the same total.
    costref = 2 * nM + chi + 3
    cost = CPCP + CS + costref + 1

    # Next are qubit costs.
    ac1 = 2 * np.ceil(np.log2(iters + 1)) - 1
    ac2 = n
    ac3 = 2 * nM
    ac47 = 5
    ac8 = beta
    ac9 = nc

    kt = 2**QR(d, m)[0]
    ac10 = m * kt + np.ceil(np.log2(d / kt))
    # This is for the equal superposition state to perform the inequality test
    # with the keep register.
    ac11 = chi
    # The qubit to control the swap of the μ and ν registers.
    ac12 = 1
    aca = ac1 + ac2 + ac3 + ac47 + ac8 + ac9 + ac11 + ac12
    # This is the data needed for the rotations.
    ac13 = beta * n / 2
    # These are the ancillas needed for adding into the phase gradient state.
    ac14 = beta - 2
    # These are the temporary ancillas in between erasing the first QROM
    # ancillas and inverting that QROM. The +m is for output of first QROM.
    acc = ac13 + ac14 + m

    if verbose:
        print("Total Toffoli cost ", cost * iters)
        print("Ancilla for first QROM ", aca + ac10)
        print("Actual ancilla ... ", np.max([aca + ac10, aca + acc]))
        print("Spacetime volume ", np.max([aca + ac10, aca + acc]) * cost)

    # First are the numbers of qubits that must be kept throughout the
    # computation. See page 18.
    if algorithm == 'half':
        # The qubits used as the control registers for the phase estimation,
        # that must be kept the whole way through. If we used independent
        # controls each time that would increase the Toffoli cost by
        # np.ceil(np.log2iters+1]]-3, while saving
        # np.ceil(np.log2iters+1]]-1 qubits
        ac1 = np.ceil(np.log2(iters + 1))
    elif algorithm == 'full':
        # The qubits used as the control registers for the phase estimation,
        # that must be kept the whole way through. If we used independent
        # controls each time that would increase the Toffoli cost by
        # np.ceil(np.log2iters+1]]-3, while saving
        # np.ceil(np.log2iters+1]]-1 qubits.
        ac1 = 2 * np.ceil(np.log2(iters + 1)) - 1
    # The system qubits that must always be included.
    ac2 = n
    # The μ and ν registers, that must be kept because they are control
    # registers that aren't fully erased and must be reflected on.
    ac3 = 2 * nM
    # These are the qubits for the spin in the control state as well as the
    # qubit that is rotated for the preparation of the equal superposition
    # state, AND the qubit that is used to control.
    # None of these are fully inversely prepared.
    ac4512 = 4
    # The qubits for the phase gradient state.
    ac8 = beta
    # This is for the equal superposition state to perform the inequality test
    # with the keep register. It must be kept around and reflected upon.
    ac11 = chi
    # The total number of permanent qubits.
    perm = ac1 + ac2 + ac3 + ac4512 + ac8 + ac11
    # In preparing the equal superposition state there are 6 temporary qubits
    # used in the rotation of the ancilla.  There are another three that are
    # needed for the temporary results of inequality tests. By far the largest
    # number, however, come from keeping the temporary ancillas from the
    # inequality tests.  That should be 3*nM+nN-4.  There are an other two
    # qubits in output at the end that will be kept until this step is undone.
    # Note: not used?
    #nN = np.ceil(np.log2(n / 2))

    # This is the maximum number of qubits used while preparing the equal
    # superposition state.
    qu1 = perm + 4 * nM - 1
    # To explain the number of temporary ancillas, we have nM+1 to perform the
    # inequality test on mu and nu with out-of-place addition.  We have another
    # nM-2 for the equality test.  Then we can do the inequality tests on
    # mu and nu with constants (temporarily) overwriting these variables, and
    # keeping nM-1 qubits on each.  Then there are another 2 temporary qubits
    # used for the reflection.  That gives 4*nM-1 total.
    # This is the number of Toffolis during this step.
    tof1 = 10 * nM + 2 * br - 9

    # This is increasing the running number of permanent ancillas by 2 for the
    #  ν=M+1 flag qubit and the success flag qubit.
    perm = perm + 2
    # The number of temporary qubits used in this computation is the same
    # as the number of Toffolis plus one.
    qu2 = perm + nM**2 + nM
    # The Toffoli cost of computing the contiguous register.
    tof2 = nM**2 + nM - 1
    # The running number of qubits is increased by the number needed for the
    # contiguous register.
    perm = perm + nc

    if algorithm == 'half':
        # Here I'm setting the k-value for the QROM by hand instead of choosing
        # the optimal one for Toffolis.
        kt = 16
    elif algorithm == 'full':
        # Here I'm setting the k-value for the QROM by hand instead of choosing
        #  the optimal one for Toffolis.
        kt = 32
    # This is the number of qubits needed during the QROM.
    qu3 = perm + m * kt + np.ceil(np.log2(d / kt))
    # The number of Toffolis for the QROM.
    tof3 = np.ceil(d / kt) + m * (kt - 1)
    # The number of ancillas used increases by the actual output size of  QROM.
    perm = perm + m
    # The number of ancilla qubits used for the subtraction for the ineql test.
    qu4 = perm + chi
    # We can use one of the qubits from the registers that are subtracted as
    # the flag qubit so we don't need an extra flag qubit.
    # The number of Toffolis needed for the inequality test. The number of
    # permanent ancillas is unchanged.*)
    tof4 = chi
    # We don't need any extra ancillas for the controlled swaps.
    qu5 = perm
    # We are swapping pairs of registers of size nM
    tof5 = 2 * nM
    # One extra ancilla for the controlled swap of mu and nu because it is
    # controlled on two qubits.
    qu6 = perm
    # One more Toffoli for the double controls.
    tof6 = nM + 1
    # Swapping based on the spin register.
    qu7 = perm
    tof7 = n / 2

    if algorithm == 'half':
        # We use these temporary ancillas for the first QROM for the rot angles.
        qu8 = perm + nM + beta * n / 4
    elif algorithm == 'full':
        # We use these temporary ancillas for the first QROM for the rot angles.
        qu8 = perm + nM + beta * n / 2
    # The cost of outputting the rotation angles including those for the
    # one-electron part.
    tof8 = M + n / 2 - 2

    if algorithm == 'half':
        # We are now need the output rotation angles, though we don't need the
        # temporary qubits from the unary iteration.
        perm = perm + beta * n / 4
    elif algorithm == 'full':
        # We are now need the output rotation angles, though we don't need the
        # temporary qubits from the unary iteration.
        perm = perm + beta * n / 2
    # We need a few temp registers for adding into the phase grad register.
    qu9 = perm + (beta - 2)

    if algorithm == 'half':
        # The cost of the rotations.
        tof9 = n * (beta - 2) / 2
        # Make a list where we keep subtr the data qubits that can be erased.
        # Table[-j*beta,{j,0,n/4-1}]+perm+(beta-2)
        qu10 = np.array([-j * beta for j in range(int(n / 4))]) \
               + perm + beta - 2
        # The cost of the rotations.
        # Table[2*(beta-2),{j,0,n/4-1}]
        tof10 = np.array([2 * (beta - 2) for j in range(int(n / 4))])
        # We've erased the data
        perm = perm - beta * n / 4
    elif algorithm == 'full':
        # The cost of the rotations.
        tof9 = n * (beta - 2)
        # Make a list where we keep subtr the data qubits that can be erased.
        # Table[-j*beta,{j,0,n/2-1}]+perm+(beta-2)
        qu10 = np.array([-j * beta for j in range(int(n / 2))]) \
               + perm + beta - 2
        # The cost of the rotations.
        # Table[2*(beta-2),{j,0,n/2-1}]
        tof10 = np.array([2 * (beta - 2) for j in range(int(n / 2))])
        # We've erased the data
        perm = perm - beta * n / 2

    # Find the k for the phase fixup for the erasure of the rotations.
    k1 = 2**QI(M + n / 2)[0]

    # Temp qubits used. Data qubits were already erased, so don't change perm.
    qu11 = perm + k1 + np.ceil(np.log2(M / k1))
    tof11 = np.ceil(M / k1) + np.ceil(n / 2 / k1) + k1

    # Swapping based on the spin register.
    qu12 = perm
    tof12 = n / 2

    # Swapping the spin registers.
    qu12a = perm
    tof12a = 1

    # Swapping based on the spin register.
    qu13 = perm
    tof13 = n / 2

    if algorithm == 'half':
        # We use these temp ancillas for the second QROM for the rot angles.
        qu14 = perm + nM - 1 + beta * n / 4
        perm = perm + beta * n / 4
    elif algorithm == 'full':
        # We use these temp ancillas for the second QROM for the rot angles.
        qu14 = perm + nM - 1 + beta * n / 2
        perm = perm + beta * n / 2
    tof14 = M - 2

    # We need a few temporary registers for adding into the phase grad register
    qu15 = perm + (beta - 2)

    if algorithm == 'half':
        # The cost of the rotations.
        tof15 = n * (beta - 2) / 2
    elif algorithm == 'full':
        # The cost of the rotations.
        tof15 = n * (beta - 2)

    # Just one Toffoli to do the controlled Z1.
    qu16 = perm
    tof16 = 1

    if algorithm == 'half':
        # Make a list where we keep subtr the data qubits that can be erased.
        # Table[-j*beta,{j,0,n/4-1}]+perm+(beta-2)
        qu17 = np.array([-j * beta for j in range(int(n / 4))
                        ]) + perm + beta - 2
        # The cost of the rotations.
        # Table[2*(beta-2),{j,0,n/4-1}]
        tof17 = np.array([2 * (beta - 2) for j in range(int(n / 4))])
        # We've erased the data.
        perm = perm - beta * n / 4
    elif algorithm == 'full':
        # Make a list where we keep subtr the data qubits that can be erased.
        # Table[-j*beta,{j,0,n/2-1}]+perm+(beta-2)
        qu17 = np.array([-j * beta for j in range(int(n / 2))]) \
               + perm + beta - 2
        # The cost of the rotations.
        # Table[2*(beta-2),{j,0,n/2-1}]
        tof17 = np.array([2 * (beta - 2) for j in range(int(n / 2))])
        # We've erased the data
        perm = perm - beta * n / 2

    # Find the k for the phase fixup for the erasure of the rotations.
    k1 = 2**QI(M)[0]

    # The temp qubits used. The data qubits were already erased,
    # so don't change perm.
    qu18 = perm + k1 + np.ceil(np.log2(M / k1))
    tof18 = np.ceil(M / k1) + k1

    # Swapping based on the spin register.
    qu19 = perm
    tof19 = n / 2

    # One extra ancilla for the controlled swap of mu and nu because it is
    # controlled on two qubits.
    qu20 = perm + 1
    # One extra Toffoli, because we are controlling on two qubits.
    tof20 = nM + 1

    # We don't need any extra ancillas for the controlled swaps.
    qu21 = perm
    # We are swapping pairs of registers of size nM
    tof21 = 2 * nM

    # The number of ancilla qubits used for the subtraction for the inequal test
    qu22 = perm + chi
    # We can use one of the qubits from the registers that are subtracted as
    # the flag qubit so we don't need an extra flag qubit.
    # The number of Toffolis needed for inverting the inequality test.
    # The number of permanent ancillas is unchanged.
    tof22 = chi
    # We can erase the data for the QROM for inverting the state preparation,
    # then do the phase fixup.
    perm = perm - m

    kt = 2**QI(d)[0]
    # This is the number of qubits needed during the QROM.
    qu23 = perm + kt + np.ceil(np.log2(d / kt))
    # The number of Toffolis for the QROM.
    tof23 = np.ceil(d / kt) + kt
    # The number of temporary qubits used in this computation is the same as
    # the number of Toffolis plus one. We are erasing the contiguous register
    # as we go so can subtract nc.
    qu24 = perm - nc + nM**2 + nM
    # The Toffoli cost of computing the contiguous register.
    tof24 = nM**2 + nM - 1
    # The contiguous register has now been deleted.
    perm = perm - nc
    # This is the maximum number of qubits used while preparing the equal
    # superposition state.
    qu25 = perm + 4 * nM - 1
    # This is the number of Toffolis during this step.
    tof25 = 10 * nM + 2 * br - 9
    # This is increasing the running number of permanent ancillas by 2 for
    # the ν=M+1 flag qubit and the success flag qubit.
    perm = perm - 2

    if algorithm == 'half':
        # We need some ancillas to perform a reflection on multiple qubits.
        # We are including one more Toffoli to make it controlled.
        qu26 = perm + costref + np.ceil(np.log2(iters + 1))
        tof26 = costref + np.ceil(np.log2(iters + 1))
    elif algorithm == 'full':
        # We need some ancillas to perform a reflection on multiple qubits.
        # We are including one more Toffoli to make it controlled.
        qu26 = perm + costref
        tof26 = costref

    qu27 = perm  # Iterate the control register.
    tof27 = 1

    # Labels
    sm = 'small element'
    pq = 'preparation QROM'
    rq = 'rotation QROM'
    ri = r'R$^{\dag}$'
    ro = 'R'
    iq = 'inverse QROM'

    color_dict = {
        sm: '#435CE8',
        pq: '#E83935',
        rq: '#F59236',
        ri: '#E3D246',
        ro: '#36B83E',
        iq: '#E83935'
    }

    if algorithm == 'half':
        tgates = np.hstack((np.array([
            tof1, tof2, tof3, tof4, tof5, tof6, tof7, tof8, tof9, tof8, tof9,
            tof9, tof8
        ]), tof10,
                            np.array([
                                tof11, tof12, tof12a, tof13, tof14, tof15,
                                tof14, tof15, tof16, tof15, tof14
                            ]), tof17,
                            np.array([
                                tof18, tof19, tof20, tof21, tof22, tof23, tof24,
                                tof25, tof26, tof27
                            ])))
        qubits = np.hstack(
            (np.array([
                qu1, qu2, qu3, qu4, qu5, qu6, qu7, qu8, qu9, qu8, qu9, qu9, qu8
            ]), qu10,
             np.array([
                 qu11, qu12, qu12a, qu13, qu14, qu15, qu14, qu15, qu16, qu15,
                 qu14
             ]), qu17,
             np.array(
                 [qu18, qu19, qu20, qu21, qu22, qu23, qu24, qu25, qu26, qu27])))
        labels = [sm, sm, pq, sm, sm, sm, sm, rq, ri, rq, ri, ro, rq] + \
                 [ro] * len(qu10) + \
                 [rq, sm, sm, sm, rq, ri, rq, ri, sm, ro, rq] + \
                 [ro] * len(qu17) + \
                 [rq, sm, sm, sm, sm, iq, sm, sm, sm, sm]

        colors = [color_dict[i] for i in labels]
    elif algorithm == 'full':
        tgates = np.hstack(
            (np.array([tof1, tof2, tof3, tof4, tof5, tof6, tof7, tof8,
                       tof9]), tof10,
             np.array([tof11, tof12, tof12a, tof13, tof14, tof15,
                       tof16]), tof17,
             np.array([
                 tof18, tof19, tof20, tof21, tof22, tof23, tof24, tof25, tof26,
                 tof27
             ])))
        qubits = np.hstack(
            (np.array([qu1, qu2, qu3, qu4, qu5, qu6, qu7, qu8, qu9]), qu10,
             np.array([qu11, qu12, qu12a, qu13, qu14, qu15, qu16]), qu17,
             np.array(
                 [qu18, qu19, qu20, qu21, qu22, qu23, qu24, qu25, qu26, qu27])))
        labels = [sm, sm, pq, sm, sm, sm, sm, rq, ri] + \
                 [ro] * len(qu10) + \
                 [rq, sm, sm, sm, rq, ri, sm] + \
                 [ro] * len(qu17) + \
                 [rq, sm, sm, sm, sm, iq, sm, sm, sm, sm]

        colors = [color_dict[i] for i in labels]

    # check lists are at least consistent
    assert all(
        len(element) == len(tgates) for element in [qubits, labels, colors])

    return tgates, qubits, labels, colors


def plot_qubit_vs_toffoli(tgates,
                          qubits,
                          labels,
                          colors,
                          tgate_label_thresh=100):
    """ Helper function to plot qubit vs toffoli similar to Figs 11 and 12 from
        'Even more efficient quantum...' paper (arXiv:2011.03494)

    Args:
        tgates (list or 1D vector) - list of toffoli values
        qubits (list or 1D vector) - list of qubit values
        labels (list) - list of labels corresponding to different steps of alg
        colors (list) - list of colors corresponding to different steps of alg
        tgate_label_thresh - don't label steps "thinner" than thresh of Toffolis
    """
    # To align the bars on the right edge pass a negative width and align='edge'
    ax = plt.gca()
    plt.bar(np.cumsum(tgates),
            qubits,
            width=-tgates,
            align='edge',
            color=colors)
    plt.bar(0, qubits[-1], width=sum(tgates), align='edge', color='#D7C4F2')
    plt.xlabel('Toffoli count')
    plt.ylabel('Number of qubits')

    # Now add the labels
    # First, group labels and neighboring tgates
    labels_grouped, tgates_grouped, qubits_grouped = group_steps(
        labels, tgates, qubits)
    for step, label in enumerate(labels_grouped):
        if 'small' in label:
            # skip the steps identified as 'small'
            continue
        elif tgates_grouped[step] < tgate_label_thresh:
            # otherwise skip really narrow steps
            continue
        else:
            x = np.cumsum(tgates_grouped)[step] - (tgates_grouped[step] * 0.5)
            y = 0.5 * (qubits_grouped[step] - qubits[-1]) + qubits[-1]
            ax.text(x,
                    y,
                    label,
                    rotation='vertical',
                    va='center',
                    ha='center',
                    fontsize='x-small')

    # Finally add system and control qubit label
    ax.text(0.5*np.sum(tgates), 0.5*qubits[-1], "System and control qubits", \
            va='center', ha='center',fontsize='x-small')

    plt.show()


def table_qubit_vs_toffoli(tgates, qubits, labels, colors):
    """ Helper function to generate qubit vs toffoli table .. text version of
         Fig 11 and Fig 12 in arXiv:2011.03494

    Args:
        tgates (list or 1D vector) - list of toffoli values
        qubits (list or 1D vector) - list of qubit values
        labels (list) - list of labels corresponding to different steps of alg
        colors (list) - list of colors corresponding to different steps of alg
    """

    print("=" * 60)
    print("{:>8s}{:>11s}{:>9s}{:>20s}{:>12s}".format('STEP', 'TOFFOLI',
                                                     'QUBIT*', 'LABEL',
                                                     'COLOR'))
    print("-" * 60)
    for step in range(len(tgates)):
        print('{:8d}{:11d}{:9d}{:>20s}{:>12s}'.format(step, int(tgates[step]),
                                                      int(qubits[step]),
                                                      labels[step],
                                                      colors[step]))
    print("=" * 60)
    print("  *Includes {:d} system and control qubits".format(int(qubits[-1])))


def group_steps(labels, tgates, qubits):
    """ Group similar adjacent steps by label. In addition to the grouped
        labels, also  returning the total Toffoli count and average qubits
        allocated for that grouping.
        Useful for collecting similar steps in the spacetime plots.

        Example:
          Input:
            labels = ['R', 'R', 'QROM', 'QROM, 'I-QROM', 'QROM', 'QROM', 'R']
            tgates = [  5,   8,     20,    10,       14,     30,     10,  20]
            qubits = [ 10,  10,     40,    20,        4,     80,     60,  60]

          Output:
            grouped_labels = ['R', 'QROM', 'I-QROM', 'QROM', 'R']
            grouped_tgates = [ 13,     30,       14,     40,  20]  (sum)
            grouped_qubits = [ 10,     30,        4,     70,  60]  (mean)

    """
    assert len(labels) == len(tgates)
    assert len(labels) == len(qubits)

    # Key function -- group identical nearest neighbors in labels (x[0])
    key_func = lambda x: x[0]

    # create grouped labels and tgates first
    grouped_labels = []
    grouped_tgates = []
    L = zip(labels, tgates)
    for label, group in itertools.groupby(L, key_func):
        grouped_labels.append(label)
        grouped_tgates.append(np.sum([i[1] for i in group]))

    # now do the grouped qubits
    # somehow doing multiple list comprehension the group breaks the grouping?
    # so we have to do this in a separate loop.
    grouped_qubits = []
    L = zip(labels, qubits)
    for label, group in itertools.groupby(L, key_func):
        grouped_qubits.append(np.mean([i[1] for i in group]))

    # sanity check -- shouldn't be losing total value in toffoli
    assert np.sum(tgates) == np.sum(grouped_tgates)
    return grouped_labels, grouped_tgates, grouped_qubits



# coverage: ignore
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from dataclasses import dataclass
from typing import Tuple
import numpy as np
import numpy.typing as npt

from openfermion.resource_estimates.pbc.thc.thc_integrals import (
    KPTHCDoubleTranslation,)
from openfermion.resource_estimates.pbc.hamiltonian import (
    HamiltonianProperties,)


@dataclass
class THCHamiltonianProperties(HamiltonianProperties):
    """Store for return values of compute_lambda function

    Extension of HamiltonianProperties dataclass to also hold the THC dimension
    (num_sym_unique).
    """

    thc_dim: int


def compute_lambda_real(
        h1: npt.NDArray,
        etaPp: npt.NDArray,
        MPQ: npt.NDArray,
        chol: npt.NDArray,
) -> Tuple[float, float, float]:
    """Compute lambda assuming real THC factors (molecular way)

    Avoids the need of a molecular object as in the molecular code.

    Args:
        h1: one-body hamiltonian
        etaPp: THC leaf tensor
        MPQ: THC central tensor.
        chol: Cholesky factors.

    Returns:
        lambda_tot: Total lambda
        lambda_one_body: One-body lambda
        lambda_two_body: Two-body lambda
    """

    # projecting into the THC basis requires each THC factor mu to be nrmlzd.
    # we roll the normalization constant into the central tensor zeta
    SPQ = etaPp.dot(
        etaPp.T)  # (nthc x norb)  x (norb x nthc) -> (nthc  x nthc) metric
    cP = np.diag(np.diag(
        SPQ))  # grab diagonal elements. equivalent to np.diag(np.diagonal(SPQ))
    # no sqrts because we have two normalized THC vectors (index by mu and nu)
    # on each side.
    MPQ_normalized = cP.dot(MPQ).dot(cP)  # get normalized zeta in Eq. 11 & 12

    lambda_z = np.sum(np.abs(MPQ_normalized)) * 0.5  # Eq. 13
    # NCR: originally Joonho's code add np.einsum('llij->ij', eri_thc)
    # NCR: I don't know how much this matters.
    T = (h1 - 0.5 * np.einsum("nil,nlj->ij", chol, chol, optimize=True) +
         np.einsum("nll,nij->ij", chol, chol, optimize=True))  # Eq. 3 + Eq. 18
    # e, v = np.linalg.eigh(T)
    e = np.linalg.eigvalsh(T)  # only need eigenvalues
    lambda_T = np.sum(
        np.abs(e))  # Eq. 19. NOTE: sum over spin orbitals removes 1/2 factor

    lambda_tot = lambda_z + lambda_T  # Eq. 20

    return lambda_tot, lambda_T, lambda_z


def compute_lambda(hcore: npt.NDArray,
                   thc_obj: KPTHCDoubleTranslation) -> HamiltonianProperties:
    """Compute one-body and two-body lambda for qubitization the THC LCU.

    Args:
        hcore: List len(kpts) long of nmo x nmo complex hermitian arrays
        thc_obj: Object of KPTHCDoubleTranslation

    Returns:
        ham_props: A HamiltonianProperties instance containing Lambda values for
            THC hamiltonian.
    """
    kpts = thc_obj.kmf.kpts
    nkpts = len(kpts)
    one_body_mat = np.empty((len(kpts)), dtype=object)
    lambda_one_body = 0.0

    for kidx in range(len(kpts)):
        # matrices for - 0.5 * sum_{Q}sum_{r}(pkrQ|rQqk)
        # and  + 0.5 sum_{Q}sum_{r}(pkqk|rQrQ)
        h1_neg = np.zeros_like(hcore[kidx])
        for qidx in range(len(kpts)):
            # - 0.5 * sum_{Q}sum_{r}(pkrQ|rQqk)
            eri_kqqk_pqrs = thc_obj.get_eri_exact([kidx, qidx, qidx, kidx])
            h1_neg -= (np.einsum("prrq->pq", eri_kqqk_pqrs, optimize=True) /
                       nkpts)

        one_body_mat[kidx] = hcore[kidx] + 0.5 * h1_neg  # + h1_pos
        one_eigs, _ = np.linalg.eigh(one_body_mat[kidx])
        lambda_one_body += np.sum(np.abs(one_eigs))

    # projecting into the THC basis requires each THC factor mu to be
    # normalized.  we roll the normalization constant into the central tensor
    # zeta # no sqrts because we have two normalized thc vectors
    # (index by mu and nu) on each side.
    norm_kP = (np.einsum("kpP,kpP->kP",
                         thc_obj.chi.conj(),
                         thc_obj.chi,
                         optimize=True)**0.5)
    lambda_two_body = 0
    for iq, zeta_Q in enumerate(thc_obj.zeta):
        # xy einsum subscript indexes G index.
        for ik in range(nkpts):
            ik_minus_q = thc_obj.k_transfer_map[iq, ik]
            gpq = thc_obj.g_mapping[iq, ik]
            for ik_prime in range(nkpts):
                ik_prime_minus_q = thc_obj.k_transfer_map[iq, ik_prime]
                gsr = thc_obj.g_mapping[iq, ik_prime]
                norm_left = norm_kP[ik] * norm_kP[ik_minus_q]
                norm_right = norm_kP[ik_prime_minus_q] * norm_kP[ik_prime]
                MPQ_normalized = (np.einsum(
                    "P,PQ,Q->PQ",
                    norm_left,
                    zeta_Q[gpq, gsr],
                    norm_right,
                    optimize=True,
                ) / nkpts)
                lambda_two_body += np.sum(np.abs(MPQ_normalized.real))
                lambda_two_body += np.sum(np.abs(MPQ_normalized.imag))
    lambda_two_body *= 2

    lambda_tot = lambda_one_body + lambda_two_body
    lambda_data = THCHamiltonianProperties(
        lambda_total=lambda_tot,
        lambda_one_body=lambda_one_body,
        lambda_two_body=lambda_two_body,
        thc_dim=thc_obj.chi.shape[-1],
    )
    return lambda_data



#coverage:ignore
""" Pretty-print a table comparing number of SF vectors versus acc and cost """
import numpy as np
import pytest

from openfermion.resource_estimates import HAVE_DEPS_FOR_RESOURCE_ESTIMATES, sf

if HAVE_DEPS_FOR_RESOURCE_ESTIMATES:
    from pyscf import scf

    from openfermion.resource_estimates.molecule import (cas_to_pyscf,
                                                         factorized_ccsd_t,
                                                         pyscf_to_cas)


@pytest.mark.skipif(not HAVE_DEPS_FOR_RESOURCE_ESTIMATES,
                    reason='pyscf and/or jax not installed.')
def generate_costing_table(pyscf_mf,
                           name='molecule',
                           rank_range=range(50, 401, 25),
                           chi=10,
                           dE=0.001,
                           use_kernel=True,
                           no_triples=False):
    """ Print a table to file for how various SF ranks impact cost, acc., etc.

    Args:
        pyscf_mf - PySCF mean field object
        name (str) - file will be saved to 'single_factorization_<name>.txt'
        rank_range (list of ints) - list number of vectors to retain in SF alg
        dE (float) - max allowable phase error (default: 0.001)
        chi (int) - number of bits for representation of coefficients
           (default: 10)
        use_kernel (bool) - re-do SCF prior to estimating CCSD(T) error?
            Will use canonical orbitals and full ERIs for the one-body
            contributions, using rank-reduced ERIs for two-body
        no_triples (bool) - if True, skip the (T) correction, doing only CCSD

    Returns:
       None
    """

    DE = dE  # max allowable phase error
    CHI = chi  # number of bits for representation of coefficients

    if isinstance(pyscf_mf, scf.rohf.ROHF):
        num_alpha, num_beta = pyscf_mf.nelec
        assert num_alpha + num_beta == pyscf_mf.mol.nelectron
    else:
        assert pyscf_mf.mol.nelectron % 2 == 0
        num_alpha = pyscf_mf.mol.nelectron // 2
        num_beta = num_alpha

    num_orb = len(pyscf_mf.mo_coeff)
    num_spinorb = num_orb * 2

    cas_info = "CAS((%sa, %sb), %so)" % (num_alpha, num_beta, num_orb)

    try:
        assert num_orb**4 == len(pyscf_mf._eri.flatten())
    except AssertionError:
        # ERIs are not in correct form in pyscf_mf._eri, so this is a quick prep
        _, pyscf_mf = cas_to_pyscf(*pyscf_to_cas(pyscf_mf))

    # Reference calculation (eri_rr = None is full rank / exact ERIs)
    escf, ecor, etot = factorized_ccsd_t(pyscf_mf,
                                         eri_rr=None,
                                         use_kernel=use_kernel,
                                         no_triples=no_triples)

    exact_etot = etot

    filename = 'single_factorization_' + name + '.txt'

    with open(filename, 'w') as f:
        print("\n Single low rank factorization data for '" + name + "'.",
              file=f)
        print("    [*] using " + cas_info, file=f)
        print("        [+]                      E(SCF): %18.8f" % escf, file=f)
        if no_triples:
            print("        [+]    Active space CCSD E(cor): %18.8f" % ecor,
                  file=f)
            print("        [+]    Active space CCSD E(tot): %18.8f" % etot,
                  file=f)
        else:
            print("        [+] Active space CCSD(T) E(cor): %18.8f" % ecor,
                  file=f)
            print("        [+] Active space CCSD(T) E(tot): %18.8f" % etot,
                  file=f)
        print("{}".format('=' * 108), file=f)
        if no_triples:
            print("{:^12} {:^18} {:^12} {:^24} {:^20} {:^20}".format('L',\
            '||ERI - SF||','lambda','CCSD error (mEh)',\
            'logical qubits', 'Toffoli count'),file=f)
        else:
            print("{:^12} {:^18} {:^12} {:^24} {:^20} {:^20}".format('L',\
            '||ERI - SF||','lambda','CCSD(T) error (mEh)',\
            'logical qubits', 'Toffoli count'),file=f)
        print("{}".format('-' * 108), file=f)
    for rank in rank_range:
        # First, up: lambda and CCSD(T)
        eri_rr, LR = sf.factorize(pyscf_mf._eri, rank)
        lam = sf.compute_lambda(pyscf_mf, LR)
        escf, ecor, etot = factorized_ccsd_t(pyscf_mf,
                                             eri_rr,
                                             use_kernel=use_kernel,
                                             no_triples=no_triples)
        error = (etot - exact_etot) * 1E3  # to mEh
        l2_norm_error_eri = np.linalg.norm(
            eri_rr - pyscf_mf._eri)  # eri reconstruction error

        # now do costing
        stps1 = sf.compute_cost(num_spinorb,
                                lam,
                                DE,
                                L=rank,
                                chi=CHI,
                                stps=20000)[0]

        _, sf_total_cost, sf_logical_qubits = sf.compute_cost(num_spinorb,
                                                              lam,
                                                              DE,
                                                              L=rank,
                                                              chi=CHI,
                                                              stps=stps1)

        with open(filename, 'a') as f:
            print(
                "{:^12} {:^18.4e} {:^12.1f} {:^24.2f} {:^20} {:^20.1e}".format(
                    rank, l2_norm_error_eri, lam, error, sf_logical_qubits,
                    sf_total_cost),
                file=f)
    with open(filename, 'a') as f:
        print("{}".format('=' * 108), file=f)



#coverage:ignore
import os
from uuid import uuid4
import scipy.optimize
import jax.numpy as jnp
from jax.config import config
from jax import jit, grad
import h5py
import numpy
import numpy.random
import numpy.linalg
from scipy.optimize import minimize
from .adagrad import adagrad

# set mkl thread count for numpy einsum/tensordot calls
# leave one CPU un used  so we can still access this computer
os.environ["MKL_NUM_THREADS"] = "{}".format(os.cpu_count() - 1)
config.update("jax_enable_x64", True)


def thc_objective_jax(xcur, norb, nthc, eri):
    """
    Loss function for THC factorization using jax numpy

    0.5 sum_{pqrs}(eri(pqrs) - G(pqrs))^{2}

    G(pqrs) = sum_{uv}X_{u,p}X_{u,q}Z_{uv}X_{v,r}X_{v,s}

    :param xcur: Current parameters for eta and Z
    :param norb: number of orbitals
    :param nthc: thc-basis dimension
    :param eri: two-electron repulsion integrals in chemist notation
    :return:
    """
    etaPp = xcur[:norb * nthc].reshape(nthc, norb)  # leaf tensor  nthc x norb
    MPQ = xcur[norb * nthc:norb * nthc + nthc * nthc].reshape(
        nthc, nthc)  # central tensor

    CprP = jnp.einsum("Pp,Pr->prP", etaPp,
                      etaPp)  # this is einsum('mp,mq->pqm', etaPp, etaPp)
    Iapprox = jnp.einsum('pqU,UV,rsV->pqrs',
                         CprP,
                         MPQ,
                         CprP,
                         optimize=[(0, 1), (0, 1)])
    deri = eri - Iapprox
    res = 0.5 * jnp.sum((deri)**2)
    return res


def thc_objective_grad_jax(xcur, norb, nthc, eri):
    """
    Gradient for THC least-squares objective jax compatible

    :param xcur: Current parameters for eta and Z
    :param norb: number of orbitals
    :param nthc: thc-basis dimension
    :param eri: two-electron repulsion integrals in chemist notation
    :param verbose: optional (False) for print iteration residual and inf norm
    """
    etaPp = xcur[:norb * nthc].reshape(nthc, norb)  # leaf tensor  nthc x norb
    MPQ = xcur[norb * nthc:norb * nthc + nthc * nthc].reshape(
        nthc, nthc)  # central tensor

    # m indexes the nthc and p,q,r,s are orbital indices
    CprP = jnp.einsum("Pp,Pr->prP", etaPp,
                      etaPp)  # this is einsum('mp,mq->pqm', etaPp, etaPp)
    Iapprox = jnp.einsum('pqU,UV,rsV->pqrs',
                         CprP,
                         MPQ,
                         CprP,
                         optimize=[(0, 1), (0, 1)])
    deri = eri - Iapprox

    # O(norb^5)
    dL_dZab = -jnp.einsum(
        'pqrs,pqA,rsB->AB', deri, CprP, CprP, optimize=[(0, 1), (0, 1)])

    # O(norb^5)
    dL_dX_GT = -2 * jnp.einsum('Tqrs,Gq,Gv,rsv->GT',
                               deri,
                               etaPp,
                               MPQ,
                               CprP,
                               optimize=[(0, 3), (1, 2), (0, 1)])

    dL_dX_GT -= 2 * jnp.einsum('pqTs,pqu,uG,Gs->GT',
                               deri,
                               CprP,
                               MPQ,
                               etaPp,
                               optimize=[(0, 1), (0, 2), (0, 1)])

    return jnp.hstack((dL_dX_GT.ravel(), dL_dZab.ravel()))


def thc_objective(xcur, norb, nthc, eri, verbose=False):
    """
    Loss function for THC factorization

    0.5 sum_{pqrs}(eri(pqrs) - G(pqrs))^{2}

    G(pqrs) = sum_{uv}X_{u,p}X_{u,q}Z_{uv}X_{v,r}X_{v,s}

    :param xcur: Current parameters for eta and Z
    :param norb: number of orbitals
    :param nthc: thc-basis dimension
    :param eri: two-electron repulsion integrals in chemist notation
    :param verbose: optional (False) for print iteration residual and inf norm
    :return:
    """
    etaPp = xcur[:norb * nthc].reshape(nthc, norb)  # leaf tensor  nthc x norb
    MPQ = xcur[norb * nthc:norb * nthc + nthc * nthc].reshape(
        nthc, nthc)  # central tensor

    CprP = numpy.einsum("Pp,Pr->prP", etaPp,
                        etaPp)  # this is einsum('mp,mq->pqm', etaPp, etaPp)
    Iapprox = numpy.einsum('pqU,UV,rsV->pqrs',
                           CprP,
                           MPQ,
                           CprP,
                           optimize=['einsum_path', (0, 1), (0, 1)])
    deri = eri - Iapprox
    res = 0.5 * numpy.sum((deri)**2)

    if verbose:
        print("res, max, lambda = {}, {}".format(res,
                                                 numpy.max(numpy.abs(deri))))

    return res


def thc_objective_regularized(xcur,
                              norb,
                              nthc,
                              eri,
                              penalty_param,
                              verbose=False):
    """
    Loss function for THC factorization

    0.5 sum_{pqrs}(eri(pqrs) - G(pqrs))^{2}

    G(pqrs) = sum_{uv}X_{u,p}X_{u,q}Z_{uv}X_{v,r}X_{v,s}

    :param xcur: Current parameters for eta and Z
    :param norb: number of orbitals
    :param nthc: thc-basis dimension
    :param eri: two-electron repulsion integrals in chemist notation
    :param verbose: optional (False) for print iteration residual and inf norm
    :return:
    """
    etaPp = xcur[:norb * nthc].reshape(nthc, norb)  # leaf tensor  nthc x norb
    MPQ = xcur[norb * nthc:norb * nthc + nthc * nthc].reshape(
        nthc, nthc)  # central tensor

    CprP = jnp.einsum("Pp,Pr->prP", etaPp,
                      etaPp)  # this is einsum('mp,mq->pqm', etaPp, etaPp)
    Iapprox = jnp.einsum('pqU,UV,rsV->pqrs',
                         CprP,
                         MPQ,
                         CprP,
                         optimize=[(0, 1), (0, 1)])
    deri = eri - Iapprox

    SPQ = etaPp.dot(
        etaPp.T)  # (nthc x norb)  x (norb x nthc) -> (nthc  x nthc) metric
    cP = jnp.diag(jnp.diag(
        SPQ))  # grab diagonal elements. equivalent to np.diag(np.diagonal(SPQ))
    # no sqrts because we have two normalized THC vectors (index by mu and nu)
    # on each side.
    MPQ_normalized = cP.dot(MPQ).dot(cP)  # get normalized zeta in Eq. 11 & 12

    lambda_z = jnp.sum(jnp.abs(MPQ_normalized)) * 0.5

    res = 0.5 * jnp.sum((deri)**2) + penalty_param * (lambda_z**2)

    if verbose:
        print("res, max, lambda**2 = {}, {}".format(res, lambda_z**2))

    return res


def thc_objective_grad(xcur, norb, nthc, eri, verbose=False):
    """
    Gradient for THC least-squares objective

    :param xcur: Current parameters for eta and Z
    :param norb: number of orbitals
    :param nthc: thc-basis dimension
    :param eri: two-electron repulsion integrals in chemist notation
    :param verbose: optional (False) for print iteration residual and inf norm
    """
    etaPp = numpy.array(xcur[:norb * nthc]).reshape(
        nthc, norb)  # leaf tensor  nthc x norb
    MPQ = numpy.array(xcur[norb * nthc:norb * nthc + nthc * nthc]).reshape(
        nthc, nthc)  # central tensor

    # m indexes the nthc and p,q,r,s are orbital indices
    CprP = numpy.einsum("Pp,Pr->prP", etaPp,
                        etaPp)  # this is einsum('mp,mq->pqm', etaPp, etaPp)
    Iapprox = numpy.einsum('pqU,UV,rsV->pqrs',
                           CprP,
                           MPQ,
                           CprP,
                           optimize=['einsum_path', (0, 1), (0, 1)])
    deri = eri - Iapprox
    res = 0.5 * numpy.sum((deri)**2)

    if verbose:
        print("res, max, lambda = {}, {}".format(res,
                                                 numpy.max(numpy.abs(deri))))

    # O(norb^5)
    dL_dZab = -numpy.einsum('pqrs,pqA,rsB->AB',
                            deri,
                            CprP,
                            CprP,
                            optimize=['einsum_path', (0, 1), (0, 1)])
    # O(norb^5)
    dL_dX_GT = -2 * numpy.einsum(
        'Tqrs,Gq,Gv,rsv->GT',
        deri,
        etaPp,
        MPQ,
        CprP,
        optimize=['einsum_path', (0, 3), (1, 2), (0, 1)])

    dL_dX_GT -= 2 * numpy.einsum(
        'pqTs,pqu,uG,Gs->GT',
        deri,
        CprP,
        MPQ,
        etaPp,
        optimize=['einsum_path', (0, 1), (0, 2), (0, 1)])

    return numpy.hstack((dL_dX_GT.ravel(), dL_dZab.ravel()))


def thc_objective_and_grad(xcur, norb, nthc, eri, verbose=False):
    """
    Loss function for THC factorization

    0.5 sum_{pqrs}(eri(pqrs) - G(pqrs))^{2}

    G(pqrs) = sum_{uv}X_{u,p}X_{u,q}Z_{uv}X_{v,r}X_{v,s}

    :param xcur: Current parameters for eta and Z
    :param norb: number of orbitals
    :param nthc: thc-basis dimension
    :param eri: two-electron repulsion integrals in chemist notation
    :param verbose: optional (False) for print iteration residual and inf norm
    :return:
    """
    etaPp = xcur[:norb * nthc].reshape(nthc, norb)  # leaf tensor  nthc x norb
    MPQ = xcur[norb * nthc:norb * nthc + nthc * nthc].reshape(
        nthc, nthc)  # central tensor
    CprP = numpy.einsum("Pp,Pr->prP", etaPp,
                        etaPp)  # this is einsum('mp,mq->pqm', etaPp, etaPp)

    Iapprox = numpy.einsum('pqU,UV,rsV->pqrs',
                           CprP,
                           MPQ,
                           CprP,
                           optimize=['einsum_path', (0, 1), (0, 1)])
    deri = eri - Iapprox
    res = 0.5 * numpy.sum((deri)**2)
    # O(norb^5)
    dL_dZab = -numpy.einsum('pqrs,pqA,rsB->AB',
                            deri,
                            CprP,
                            CprP,
                            optimize=['einsum_path', (0, 1), (0, 1)])
    # O(norb^4 * nthc)
    dL_dX_GT = -2 * numpy.einsum(
        'Tqrs,Gq,Gv,rsv->GT',
        deri,
        etaPp,
        MPQ,
        CprP,
        optimize=['einsum_path', (0, 3), (1, 2), (0, 1)])

    dL_dX_GT -= 2 * numpy.einsum(
        'pqTs,pqu,uG,Gs->GT',
        deri,
        CprP,
        MPQ,
        etaPp,
        optimize=['einsum_path', (0, 1), (0, 2), (0, 1)])

    return res, numpy.hstack((dL_dX_GT.ravel(), dL_dZab.ravel()))


def cp_ls_cholesky_factor_objective(beta_gamma,
                                    norb,
                                    nthc,
                                    cholesky_factor,
                                    calcgrad=False):
    """cholesky_factor is reshaped into (norb, norb, num_cholesky)

    Cholesky factor B_{ab,x}

    Lst sq fit obj ||B_{ab,x} - sum_{r}beta_{a,x}beta_{b,x}gamma_{ab,x}||

    This function provides the objective function value and gradient with
      respect to beta and gamma
    """
    # compute objective
    num_cholfactors = cholesky_factor.shape[-1]
    beta_bR = beta_gamma[:norb * nthc].reshape((norb, nthc))
    gamma_yR = beta_gamma[norb * nthc:norb * nthc +
                          nthc * num_cholfactors].reshape(
                              (num_cholfactors, nthc))
    beta_abR = numpy.einsum('aR,bR->abR', beta_bR, beta_bR)
    chol_approx = numpy.einsum('abR,XR->abX', beta_abR, gamma_yR)
    delta = cholesky_factor - chol_approx
    fval = 0.5 * numpy.sum((delta)**2)

    if calcgrad:
        # compute grad
        # \partial O / \partial beta_{c,s}
        grad_beta = -2 * numpy.einsum('Cbx,bS,xS->CS',
                                      delta,
                                      beta_bR,
                                      gamma_yR,
                                      optimize=['einsum_path', (0, 2), (0, 1)])
        grad_gamma = -numpy.einsum('abY,aS,bS->YS',
                                   delta,
                                   beta_bR,
                                   beta_bR,
                                   optimize=['einsum_path', (1, 2), (0, 1)])
        grad = numpy.hstack((grad_beta.ravel(), grad_gamma.ravel()))
        return fval, grad
    else:
        return fval



# coverage: ignore
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from functools import reduce
import numpy as np
import pytest

from openfermion.resource_estimates import HAVE_DEPS_FOR_RESOURCE_ESTIMATES

if HAVE_DEPS_FOR_RESOURCE_ESTIMATES:
    from pyscf.pbc import mp

    from openfermion.resource_estimates.pbc.df.compute_lambda_df import (
        compute_lambda,)
    from openfermion.resource_estimates.pbc.df.df_integrals import (
        DFABKpointIntegrals,)
    from openfermion.resource_estimates.pbc.hamiltonian import (
        cholesky_from_df_ints,)
    from openfermion.resource_estimates.pbc.testing import (
        make_diamond_113_szv,)


@pytest.mark.skipif(not HAVE_DEPS_FOR_RESOURCE_ESTIMATES,
                    reason='pyscf and/or jax not installed.')
def test_lambda_calc():
    mf = make_diamond_113_szv()
    mymp = mp.KMP2(mf)
    Luv = cholesky_from_df_ints(mymp)
    helper = DFABKpointIntegrals(cholesky_factor=Luv, kmf=mf)
    helper.double_factorize(thresh=1.0e-13)

    hcore_ao = mf.get_hcore()
    hcore_mo = np.asarray([
        reduce(np.dot, (mo.T.conj(), hcore_ao[k], mo))
        for k, mo in enumerate(mf.mo_coeff)
    ])

    lambda_data = compute_lambda(hcore_mo, helper)
    assert np.isclose(lambda_data.lambda_total, 179.62240330857406)

    lambda_two_body = 0
    lambda_two_body_v2 = 0
    nkpts = len(mf.kpts)
    for qidx in range(nkpts):
        aval_to_square = np.zeros((helper.naux), dtype=np.complex128)
        bval_to_square = np.zeros((helper.naux), dtype=np.complex128)

        aval_to_square_v2 = np.zeros((helper.naux), dtype=np.complex128)
        bval_to_square_v2 = np.zeros((helper.naux), dtype=np.complex128)

        for kidx in range(nkpts):
            Amats, Bmats = helper.build_A_B_n_q_k_from_chol(qidx, kidx)
            Amats /= np.sqrt(nkpts)
            Bmats /= np.sqrt(nkpts)
            wa, _ = np.linalg.eigh(Amats)
            wb, _ = np.linalg.eigh(Bmats)
            aval_to_square += np.einsum("npq->n", np.abs(Amats)**2)
            bval_to_square += np.einsum("npq->n", np.abs(Bmats)**2)

            aval_to_square_v2 += np.sum(np.abs(wa)**2, axis=-1)
            bval_to_square_v2 += np.sum(np.abs(wb)**2, axis=-1)
            assert np.allclose(
                np.sum(np.abs(wa)**2, axis=-1),
                np.einsum("npq->n",
                          np.abs(Amats)**2),
            )

        lambda_two_body += np.sum(aval_to_square)
        lambda_two_body += np.sum(bval_to_square)

        lambda_two_body_v2 += np.sum(aval_to_square_v2)
        lambda_two_body_v2 += np.sum(bval_to_square_v2)

    assert np.isclose(lambda_two_body, lambda_two_body_v2)



#coverage:ignore
""" Compute lambda for double low rank factoriz. method of von Burg, et al """
import numpy as np
from openfermion.resource_estimates.molecule import pyscf_to_cas


def compute_lambda(pyscf_mf, df_factors):
    """ Compute lambda for Hamiltonian using DF method of von Burg, et al.

    Args:
        pyscf_mf - Pyscf mean field object
        df_factors (ndarray) - (N x N x rank) array of DF factors from ERI

    Returns:
        lambda_tot (float) - lambda value for the single factorized Hamiltonian
    """
    # grab tensors from pyscf_mf object
    h1, eri_full, _, _, _ = pyscf_to_cas(pyscf_mf)

    # one body contributions
    T = h1 - 0.5 * np.einsum("illj->ij", eri_full) + np.einsum(
        "llij->ij", eri_full)
    e, _ = np.linalg.eigh(T)
    lambda_T = np.sum(np.abs(e))

    # two body contributions
    lambda_F = 0.0
    for vector in range(df_factors.shape[2]):
        Lij = df_factors[:, :, vector]
        #e, v = np.linalg.eigh(Lij)
        e = np.linalg.eigvalsh(Lij)  # just need eigenvals
        lambda_F += 0.25 * np.sum(np.abs(e))**2

    lambda_tot = lambda_T + lambda_F

    return lambda_tot



#coverage:ignore
"""
Compute lambdas for THC according to
PRX QUANTUM 2, 030305 (2021) Section II. D.
"""
import numpy as np
from openfermion.resource_estimates.molecule import pyscf_to_cas


def compute_lambda(pyscf_mf,
                   etaPp: np.ndarray,
                   MPQ: np.ndarray,
                   use_eri_thc_for_t=False):
    """
    Compute lambda thc

    Args:
        pyscf_mf - PySCF mean field object
        etaPp - leaf tensor for THC that is dim(nthc x norb).  The nthc and norb
                is inferred from this quantity.
        MPQ - central tensor for THC factorization. dim(nthc x nthc)

    Returns:
    """

    nthc = etaPp.shape[0]

    # grab tensors from pyscf_mf object
    h1, eri_full, _, _, _ = pyscf_to_cas(pyscf_mf)

    # computing Least-squares THC residual
    CprP = np.einsum("Pp,Pr->prP", etaPp,
                     etaPp)  # this is einsum('mp,mq->pqm', etaPp, etaPp)
    BprQ = np.tensordot(CprP, MPQ, axes=([2], [0]))
    Iapprox = np.tensordot(CprP, np.transpose(BprQ), axes=([2], [0]))
    deri = eri_full - Iapprox
    res = 0.5 * np.sum((deri)**2)

    # NOTE: remove in future once we resolve why it was used in the first place.
    # NOTE: see T construction for details.
    eri_thc = np.einsum("Pp,Pr,Qq,Qs,PQ->prqs",
                        etaPp,
                        etaPp,
                        etaPp,
                        etaPp,
                        MPQ,
                        optimize=True)

    # projecting into the THC basis requires each THC factor mu to be nrmlzd.
    # we roll the normalization constant into the central tensor zeta
    SPQ = etaPp.dot(
        etaPp.T)  # (nthc x norb)  x (norb x nthc) -> (nthc  x nthc) metric
    cP = np.diag(np.diag(
        SPQ))  # grab diagonal elements. equivalent to np.diag(np.diagonal(SPQ))
    # no sqrts because we have two normalized THC vectors (index by mu and nu)
    # on each side.
    MPQ_normalized = cP.dot(MPQ).dot(cP)  # get normalized zeta in Eq. 11 & 12

    lambda_z = np.sum(np.abs(MPQ_normalized)) * 0.5  # Eq. 13
    # NCR: originally Joonho's code add np.einsum('llij->ij', eri_thc)
    # NCR: I don't know how much this matters.
    if use_eri_thc_for_t:
        # use eri_thc for second coulomb contraction.  This was in the original
        # code which is different than what the paper says.
        T = h1 - 0.5 * np.einsum("illj->ij", eri_full) + np.einsum(
            "llij->ij", eri_thc)  # Eq. 3 + Eq. 18
    else:
        T = h1 - 0.5 * np.einsum("illj->ij", eri_full) + np.einsum(
            "llij->ij", eri_full)  # Eq. 3 + Eq. 18
    #e, v = np.linalg.eigh(T)
    e = np.linalg.eigvalsh(T)  # only need eigenvalues
    lambda_T = np.sum(
        np.abs(e))  # Eq. 19. NOTE: sum over spin orbitals removes 1/2 factor

    lambda_tot = lambda_z + lambda_T  # Eq. 20

    #return nthc, np.sqrt(res), res, lambda_T, lambda_z, lambda_tot
    return lambda_tot, nthc, np.sqrt(res), res, lambda_T, lambda_z



# coverage: ignore
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import numpy as np

import pytest

from openfermion.resource_estimates import HAVE_DEPS_FOR_RESOURCE_ESTIMATES

if HAVE_DEPS_FOR_RESOURCE_ESTIMATES:
    from openfermion.resource_estimates.pbc import sparse
    from openfermion.resource_estimates.pbc.testing import (
        make_diamond_113_szv,)


@pytest.mark.skipif(not HAVE_DEPS_FOR_RESOURCE_ESTIMATES,
                    reason='pyscf and/or jax not installed.')
def test_generate_costing_table_sparse():
    mf = make_diamond_113_szv()
    thresh = np.array([1e-1, 1e-2, 1e-12])
    table = sparse.generate_costing_table(mf,
                                          thresholds=thresh,
                                          chi=17,
                                          dE_for_qpe=1e-3)
    assert np.allclose(table.dE, 1e-3)
    assert np.allclose(table.chi, 17)
    assert np.allclose(table.cutoff, thresh)
    assert np.isclose(table.approx_energy.values[2],
                      table.exact_energy.values[0])
    assert not np.isclose(table.approx_energy.values[0],
                          table.exact_energy.values[0])



#coverage:ignore
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from openfermion.resource_estimates import HAVE_DEPS_FOR_RESOURCE_ESTIMATES

if HAVE_DEPS_FOR_RESOURCE_ESTIMATES:
    from .pyscf_utils import (avas_active_space, cas_to_pyscf, ccsd_t,
                              factorized_ccsd_t, get_num_active_alpha_beta,
                              load_casfile_to_pyscf, localize, open_shell_t1_d1,
                              pyscf_to_cas, save_pyscf_to_casfile, stability)



# coverage: ignore
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import itertools

import numpy as np
import pytest

from openfermion.resource_estimates import HAVE_DEPS_FOR_RESOURCE_ESTIMATES

if HAVE_DEPS_FOR_RESOURCE_ESTIMATES:
    from pyscf.pbc import cc, mp

    from openfermion.resource_estimates.pbc.hamiltonian import (
        build_hamiltonian, cholesky_from_df_ints)
    from openfermion.resource_estimates.pbc.testing import make_diamond_113_szv


@pytest.mark.skipif(not HAVE_DEPS_FOR_RESOURCE_ESTIMATES,
                    reason='pyscf and/or jax not installed.')
def test_build_hamiltonian():
    mf = make_diamond_113_szv()
    nmo = mf.mo_coeff[0].shape[-1]
    naux = 108
    hcore, chol = build_hamiltonian(mf)
    nkpts = len(mf.mo_coeff)
    assert hcore.shape == (nkpts, nmo, nmo)
    assert chol.shape == (nkpts, nkpts)
    assert chol[0, 0].shape == (naux, nmo, nmo)


@pytest.mark.skipif(not HAVE_DEPS_FOR_RESOURCE_ESTIMATES,
                    reason='pyscf and/or jax not installed.')
def test_pyscf_chol_from_df():
    mf = make_diamond_113_szv()
    mymp = mp.KMP2(mf)
    nmo = mymp.nmo
    nocc = mymp.nocc
    nvir = nmo - nocc
    Luv = cholesky_from_df_ints(mymp)

    # 1. Test that the DF integrals give the correct SCF energy (oo block)
    mf.exxdiv = None  # exclude ewald exchange correction
    Eref = mf.energy_elec()[1]
    Eout = 0.0j
    nkpts = len(mf.mo_coeff)
    for ik, jk in itertools.product(range(nkpts), repeat=2):
        Lii = Luv[ik, ik][:, :nocc, :nocc]
        Ljj = Luv[jk, jk][:, :nocc, :nocc]
        Lij = Luv[ik, jk][:, :nocc, :nocc]
        Lji = Luv[jk, ik][:, :nocc, :nocc]
        oooo_d = np.einsum("Lij,Lkl->ijkl", Lii, Ljj) / nkpts
        oooo_x = np.einsum("Lij,Lkl->ijkl", Lij, Lji) / nkpts
        Eout += 2.0 * np.einsum("iijj->", oooo_d)
        Eout -= np.einsum("ijji->", oooo_x)
    assert abs(Eout / nkpts - Eref) < 1e-12

    # 2. Test that the DF integrals agree with those from MP2 (ov block)
    from pyscf.pbc.mp.kmp2 import _init_mp_df_eris

    Ltest = _init_mp_df_eris(mymp)
    for ik, jk in itertools.product(range(nkpts), repeat=2):
        assert np.allclose(Luv[ik, jk][:, :nocc, nocc:],
                           Ltest[ik, jk],
                           atol=1e-12)

    # 3. Test that the DF integrals have correct vvvv block (vv)
    Ivvvv = np.zeros((nkpts, nkpts, nkpts, nvir, nvir, nvir, nvir),
                     dtype=np.complex128)
    for ik, jk, kk in itertools.product(range(nkpts), repeat=3):
        lk = mymp.khelper.kconserv[ik, jk, kk]
        Lij = Luv[ik, jk][:, nocc:, nocc:]
        Lkl = Luv[kk, lk][:, nocc:, nocc:]
        Imo = np.einsum("Lij,Lkl->ijkl", Lij, Lkl)
        Ivvvv[ik, jk, kk] = Imo / nkpts

    mycc = cc.KRCCSD(mf)
    eris = mycc.ao2mo()
    assert np.allclose(eris.vvvv,
                       Ivvvv.transpose(0, 2, 1, 3, 5, 4, 6),
                       atol=1e-12)



# coverage: ignore
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import numpy as np
import pytest

from openfermion.resource_estimates import HAVE_DEPS_FOR_RESOURCE_ESTIMATES

if HAVE_DEPS_FOR_RESOURCE_ESTIMATES:
    from ase.build import bulk
    from pyscf.pbc import gto, scf
    from pyscf.pbc.dft import numint
    from pyscf.pbc.lib.kpts_helper import get_kconserv, member, unique
    from pyscf.pbc.tools import pyscf_ase

    from openfermion.resource_estimates.pbc.hamiltonian import \
        build_momentum_transfer_mapping
    from openfermion.resource_estimates.pbc.thc.factorizations.isdf import (
        build_eri_isdf_double_translation, build_eri_isdf_single_translation,
        build_g_vector_mappings_double_translation,
        build_g_vector_mappings_single_translation, build_g_vectors,
        build_kpoint_zeta, build_minus_q_g_mapping, get_miller,
        inverse_g_map_double_translation, solve_kmeans_kpisdf, solve_qrcp_isdf,
        supercell_isdf)
    from openfermion.resource_estimates.pbc.thc.factorizations.kmeans import \
        KMeansCVT


@pytest.mark.skipif(not HAVE_DEPS_FOR_RESOURCE_ESTIMATES,
                    reason='pyscf and/or jax not installed.')
def test_supercell_isdf_gamma():
    cell = gto.Cell()
    cell.atom = """
    C 0.000000000000   0.000000000000   0.000000000000
    C 1.685068664391   1.685068664391   1.685068664391
    """
    cell.basis = "gth-szv"
    cell.pseudo = "gth-hf-rev"
    cell.a = """
    0.000000000, 3.370137329, 3.370137329
    3.370137329, 0.000000000, 3.370137329
    3.370137329, 3.370137329, 0.000000000"""
    cell.unit = "B"
    cell.verbose = 0
    cell.mesh = [11] * 3
    cell.build(parse_arg=False)
    # kpts = cell.make_kpts(kmesh, scaled_center=[0.2, 0.3, 0.5])

    mf = scf.RHF(cell)
    mf.kernel()

    from pyscf.pbc.dft import gen_grid

    grid_inst = gen_grid.UniformGrids(cell)
    grid_points = cell.gen_uniform_grids(mf.with_df.mesh)
    orbitals = numint.eval_ao(cell, grid_points)
    orbitals_mo = np.einsum("Rp,pi->Ri", orbitals, mf.mo_coeff, optimize=True)
    num_mo = mf.mo_coeff.shape[1]
    num_interp_points = np.prod(cell.mesh)
    interp_indx = np.arange(np.prod(cell.mesh))

    chi, zeta, Theta = supercell_isdf(mf.with_df,
                                      interp_indx,
                                      orbitals=orbitals_mo,
                                      grid_points=grid_points)
    assert Theta.shape == (len(grid_points), num_interp_points)
    # Check overlap
    # Evaluate overlap from orbitals. Do it integral way to ensure
    # discretization error is the same from using coarse FFT grid for testing
    # speed
    ovlp_ao = np.einsum(
        "mp,mq,m->pq",
        orbitals.conj(),
        orbitals,
        grid_inst.weights,
        optimize=True,
    )
    ovlp_mo = np.einsum("pi,pq,qj->ij",
                        mf.mo_coeff.conj(),
                        ovlp_ao,
                        mf.mo_coeff,
                        optimize=True)
    ovlp_mu = np.einsum("Rm,R->m", Theta, grid_inst.weights, optimize=True)
    orbitals_mo_interp = orbitals_mo[interp_indx]
    ovlp_isdf = np.einsum(
        "mi,mj,m->ij",
        orbitals_mo_interp.conj(),
        orbitals_mo_interp,
        ovlp_mu,
        optimize=True,
    )
    assert np.allclose(ovlp_mo, ovlp_isdf)
    # Check ERIs.
    eri_ref = mf.with_df.ao2mo(mf.mo_coeff)
    from pyscf import ao2mo

    eri_ref = ao2mo.restore(1, eri_ref, num_mo)
    # THC eris
    Lijn = np.einsum("mi,mj,mn->ijn", chi.conj(), chi, zeta, optimize=True)
    eri_thc = np.einsum("ijn,nk,nl->ijkl", Lijn, chi.conj(), chi, optimize=True)
    assert np.allclose(eri_thc, eri_ref)


@pytest.mark.skipif(not HAVE_DEPS_FOR_RESOURCE_ESTIMATES,
                    reason='pyscf and/or jax not installed.')
def test_supercell_isdf_complex():
    cell = gto.Cell()
    cell.atom = """
    C 0.000000000000   0.000000000000   0.000000000000
    C 1.685068664391   1.685068664391   1.685068664391
    """
    cell.basis = "gth-szv"
    cell.pseudo = "gth-hf-rev"
    cell.a = """
    0.000000000, 3.370137329, 3.370137329
    3.370137329, 0.000000000, 3.370137329
    3.370137329, 3.370137329, 0.000000000"""
    cell.unit = "B"
    cell.mesh = [11] * 3
    cell.verbose = 0
    cell.build(parse_arg=False)
    # kpts = cell.make_kpts(kmesh, scaled_center=[0.2, 0.3, 0.5])

    mf = scf.RHF(cell, kpt=np.array([0.1, -0.001, 0.022]))
    mf.kernel()
    assert np.max(np.abs(mf.mo_coeff.imag)) > 1e-2

    from pyscf.pbc.dft import gen_grid

    grid_inst = gen_grid.UniformGrids(cell)
    grid_points = cell.gen_uniform_grids(mf.with_df.mesh)
    orbitals = numint.eval_ao(cell, grid_points, kpt=mf.kpt)
    orbitals_mo = np.einsum("Rp,pi->Ri", orbitals, mf.mo_coeff, optimize=True)
    num_mo = mf.mo_coeff.shape[1]
    num_interp_points = 10 * num_mo
    nocc = cell.nelec[0]
    density = np.einsum(
        "Ri,Ri->R",
        orbitals_mo[:, :nocc].conj(),
        orbitals_mo[:, :nocc],
        optimize=True,
    )
    kmeans = KMeansCVT(grid_points, max_iteration=500)
    interp_indx = kmeans.find_interpolating_points(
        num_interp_points,
        density.real,
        verbose=False,
    )

    chi, zeta, Theta = supercell_isdf(mf.with_df,
                                      interp_indx,
                                      orbitals=orbitals_mo,
                                      grid_points=grid_points)
    assert Theta.shape == (len(grid_points), num_interp_points)
    # Check overlap
    # Evaluate overlap from orbitals. Do it integral way to ensure
    # discretization error is the same from using coarse FFT grid for testing
    # speed
    ovlp_ao = np.einsum(
        "mp,mq,m->pq",
        orbitals.conj(),
        orbitals,
        grid_inst.weights,
        optimize=True,
    )
    ovlp_mo = np.einsum("pi,pq,qj->ij",
                        mf.mo_coeff.conj(),
                        ovlp_ao,
                        mf.mo_coeff,
                        optimize=True)
    ovlp_mu = np.einsum("Rm,R->m", Theta, grid_inst.weights, optimize=True)
    orbitals_mo_interp = orbitals_mo[interp_indx]
    ovlp_isdf = np.einsum(
        "mi,mj,m->ij",
        orbitals_mo_interp.conj(),
        orbitals_mo_interp,
        ovlp_mu,
        optimize=True,
    )
    assert np.allclose(ovlp_mo, ovlp_isdf)
    # Check ERIs.
    eri_ref = mf.with_df.ao2mo(mf.mo_coeff, kpts=mf.kpt.reshape((1, -1)))
    # Check there is a complex component
    assert np.max(np.abs(eri_ref.imag)) > 1e-3
    # for complex integrals ao2mo will yield num_mo^4 elements.
    eri_ref = eri_ref.reshape((num_mo,) * 4)
    # THC eris
    Lijn = np.einsum("mi,mj,mn->ijn", chi.conj(), chi, zeta, optimize=True)
    eri_thc = np.einsum("ijn,nk,nl->ijkl", Lijn, chi.conj(), chi, optimize=True)
    assert np.allclose(eri_thc, eri_ref)


@pytest.mark.skipif(not HAVE_DEPS_FOR_RESOURCE_ESTIMATES,
                    reason='pyscf and/or jax not installed.')
def test_G_vector_mapping_double_translation():
    ase_atom = bulk("AlN", "wurtzite", a=3.11, c=4.98)
    cell = gto.Cell()
    cell.exp_to_discard = 0.1
    cell.atom = pyscf_ase.ase_atoms_to_pyscf(ase_atom)
    cell.a = ase_atom.cell[:].copy()
    cell.basis = "gth-szv"
    cell.pseudo = "gth-hf-rev"
    cell.verbose = 0
    cell.build(parse_arg=False)

    nk = 3
    kmesh = [nk, nk, nk]
    kpts = cell.make_kpts(kmesh)

    momentum_map = build_momentum_transfer_mapping(cell, kpts)
    (
        G_vecs,
        G_map,
        G_unique,
        _,
    ) = build_g_vector_mappings_double_translation(cell, kpts, momentum_map)
    num_kpts = len(kpts)
    for iq in range(num_kpts):
        for ikp in range(num_kpts):
            ikq = momentum_map[iq, ikp]
            q = kpts[ikp] - kpts[ikq]
            G_shift = G_vecs[G_map[iq, ikp]]
            assert np.allclose(q, kpts[iq] + G_shift)
    for iq in range(num_kpts):
        unique_G = np.unique(G_map[iq])
        for i, G in enumerate(G_map[iq]):
            assert unique_G[G_unique[iq][i]] == G

    inv_G_map = inverse_g_map_double_translation(cell, kpts, momentum_map)
    for iq in range(num_kpts):
        for ik in range(num_kpts):
            ix_G_qk = G_map[iq, ik]
            assert ik in inv_G_map[iq, ix_G_qk]


@pytest.mark.skipif(not HAVE_DEPS_FOR_RESOURCE_ESTIMATES,
                    reason='pyscf and/or jax not installed.')
def test_G_vector_mapping_single_translation():
    cell = gto.Cell()
    cell.atom = """
    C 0.000000000000   0.000000000000   0.000000000000
    C 1.685068664391   1.685068664391   1.685068664391
    """
    cell.basis = "gth-szv"
    cell.pseudo = "gth-hf-rev"
    cell.a = """
    0.000000000, 3.370137329, 3.370137329
    3.370137329, 0.000000000, 3.370137329
    3.370137329, 3.370137329, 0.000000000"""
    cell.unit = "B"
    cell.verbose = 0
    cell.build(parse_arg=False)

    nk = 3
    kmesh = [nk, nk, nk]
    kpts = cell.make_kpts(kmesh)
    num_kpts = len(kpts)

    kpts_pq = np.array([(kp, kpts[ikq])
                        for ikp, kp in enumerate(kpts)
                        for ikq in range(num_kpts)])

    kpts_pq_indx = np.array([
        (ikp, ikq) for ikp, kp in enumerate(kpts) for ikq in range(num_kpts)
    ])
    transfers = kpts_pq[:, 0] - kpts_pq[:, 1]
    assert len(transfers) == (nk**3)**2
    _, unique_indx, _ = unique(transfers)
    (
        _,
        _,
        G_unique,
        delta_Gs,
    ) = build_g_vector_mappings_single_translation(cell, kpts,
                                                   kpts_pq_indx[unique_indx])
    kconserv = get_kconserv(cell, kpts)
    for ikp in range(num_kpts):
        for ikq in range(num_kpts):
            for ikr in range(num_kpts):
                iks = kconserv[ikp, ikq, ikr]
                delta_G_expected = kpts[ikp] - kpts[ikq] + kpts[ikr] - kpts[iks]
                q = kpts[ikp] - kpts[ikq]
                qindx = member(q, transfers[unique_indx])[0]
                # print(q, len(transfers[unique_indx]), len(transfers))
                dG_indx = G_unique[qindx, ikr]
                # print(qindx, dG_indx, delta_Gs[qindx].shape)
                # print(qindx)
                # print(len(delta_Gs[qindx]), dG_indx)
                delta_G = delta_Gs[qindx][dG_indx]
                assert np.allclose(delta_G_expected, delta_G)


@pytest.mark.skipif(not HAVE_DEPS_FOR_RESOURCE_ESTIMATES,
                    reason='pyscf and/or jax not installed.')
def test_kpoint_isdf_double_translation():
    cell = gto.Cell()
    cell.atom = """
    C 0.000000000000   0.000000000000   0.000000000000
    C 1.685068664391   1.685068664391   1.685068664391
    """
    cell.basis = "gth-szv"
    cell.pseudo = "gth-hf-rev"
    cell.a = """
    0.000000000, 3.370137329, 3.370137329
    3.370137329, 0.000000000, 3.370137329
    3.370137329, 3.370137329, 0.000000000"""
    cell.unit = "B"
    cell.verbose = 0
    cell.mesh = [11] * 3  # set to coarse value to just check ERIS numerically.
    cell.build(parse_arg=False)

    kmesh = [1, 2, 1]
    kpts = cell.make_kpts(kmesh)
    mf = scf.KRHF(cell, kpts)
    mf.kernel()

    momentum_map = build_momentum_transfer_mapping(cell, kpts)
    num_mo = mf.mo_coeff[0].shape[-1]
    num_thc = np.prod(cell.mesh)
    kpt_thc = solve_kmeans_kpisdf(
        mf,
        num_thc,
        use_density_guess=True,
        verbose=False,
        single_translation=False,
    )
    num_kpts = len(momentum_map)
    for iq in range(1, num_kpts):
        for ikp in range(num_kpts):
            ikq = momentum_map[iq, ikp]
            for iks in range(num_kpts):
                ikr = momentum_map[iq, iks]
                _ = kpt_thc.g_mapping[iq, iks]
                kpt_pqrs = [kpts[ikp], kpts[ikq], kpts[ikr], kpts[iks]]
                mos_pqrs = [
                    mf.mo_coeff[ikp],
                    mf.mo_coeff[ikq],
                    mf.mo_coeff[ikr],
                    mf.mo_coeff[iks],
                ]
                eri_pqrs = mf.with_df.ao2mo(mos_pqrs, kpt_pqrs,
                                            compact=False).reshape(
                                                (num_mo,) * 4)
                eri_pqrs_isdf = build_eri_isdf_double_translation(
                    kpt_thc.chi,
                    kpt_thc.zeta,
                    iq,
                    [ikp, ikq, ikr, iks],
                    kpt_thc.g_mapping,
                )
                assert np.allclose(eri_pqrs, eri_pqrs_isdf)


@pytest.mark.skipif(not HAVE_DEPS_FOR_RESOURCE_ESTIMATES,
                    reason='pyscf and/or jax not installed.')
def test_kpoint_isdf_single_translation():
    cell = gto.Cell()
    cell.atom = """
    C 0.000000000000   0.000000000000   0.000000000000
    C 1.685068664391   1.685068664391   1.685068664391
    """
    cell.basis = "gth-szv"
    cell.pseudo = "gth-hf-rev"
    cell.a = """
    0.000000000, 3.370137329, 3.370137329
    3.370137329, 0.000000000, 3.370137329
    3.370137329, 3.370137329, 0.000000000"""
    cell.unit = "B"
    cell.verbose = 0
    cell.mesh = [11] * 3  # set to coarse value to just check ERIS numerically.
    cell.build(parse_arg=False)

    kmesh = [1, 2, 1]
    kpts = cell.make_kpts(kmesh)
    mf = scf.KRHF(cell, kpts)
    mf.kernel()

    num_thc = np.prod(cell.mesh)
    num_kpts = len(kpts)
    kpt_thc = solve_kmeans_kpisdf(
        mf,
        num_thc,
        use_density_guess=True,
        verbose=False,
        single_translation=True,
    )
    kpts_pq = np.array([(kp, kpts[ikq])
                        for ikp, kp in enumerate(kpts)
                        for ikq in range(num_kpts)])
    transfers = kpts_pq[:, 0] - kpts_pq[:, 1]
    _, unique_indx, _ = unique(transfers)
    kconserv = get_kconserv(cell, kpts)
    num_mo = mf.mo_coeff[0].shape[-1]
    for ikp in range(num_kpts):
        for ikq in range(num_kpts):
            for ikr in range(num_kpts):
                iks = kconserv[ikp, ikq, ikr]
                kpt_pqrs = [kpts[ikp], kpts[ikq], kpts[ikr], kpts[iks]]
                mos_pqrs = [
                    mf.mo_coeff[ikp],
                    mf.mo_coeff[ikq],
                    mf.mo_coeff[ikr],
                    mf.mo_coeff[iks],
                ]
                eri_pqrs = mf.with_df.ao2mo(mos_pqrs, kpt_pqrs,
                                            compact=False).reshape(
                                                (num_mo,) * 4)
                q = kpts[ikp] - kpts[ikq]
                qindx = member(q, transfers[unique_indx])[0]
                eri_pqrs_isdf = build_eri_isdf_single_translation(
                    kpt_thc.chi,
                    kpt_thc.zeta,
                    qindx,
                    [ikp, ikq, ikr, iks],
                    kpt_thc.g_mapping,
                )
                assert np.allclose(eri_pqrs, eri_pqrs_isdf)


def get_complement(miller_indx, kmesh):
    complement = ~miller_indx
    complement[np.where(np.array(kmesh) == 1)] = 0
    return complement


@pytest.mark.skipif(not HAVE_DEPS_FOR_RESOURCE_ESTIMATES,
                    reason='pyscf and/or jax not installed.')
@pytest.mark.slow
def test_kpoint_isdf_symmetries():
    cell = gto.Cell()
    cell.atom = """
    C 0.000000000000   0.000000000000   0.000000000000
    C 1.685068664391   1.685068664391   1.685068664391
    """
    cell.basis = "gth-szv"
    cell.pseudo = "gth-hf-rev"
    cell.a = """
    0.000000000, 3.370137329, 3.370137329
    3.370137329, 0.000000000, 3.370137329
    3.370137329, 3.370137329, 0.000000000"""
    cell.unit = "B"
    cell.mesh = [11] * 3
    cell.verbose = 0
    cell.build(parse_arg=False)

    kmesh = [1, 2, 3]
    kpts = cell.make_kpts(kmesh)
    mf = scf.KRHF(cell, kpts)
    mf.kernel()
    num_thc = np.prod(
        cell.mesh)  # should be no THC error when selecting all the grid points.
    kpt_thc = solve_kmeans_kpisdf(
        mf,
        num_thc,
        use_density_guess=True,
        verbose=False,
        single_translation=False,
    )
    momentum_map = build_momentum_transfer_mapping(cell, kpts)
    (
        _,
        _,
        G_unique,
        delta_Gs,
    ) = build_g_vector_mappings_double_translation(cell, kpts, momentum_map)
    _, minus_Q_G_map_unique = build_minus_q_g_mapping(cell, kpts, momentum_map)
    num_kpts = len(kpts)
    # Test symmetries from Appendix D of https://arxiv.org/pdf/2302.05531.pdf
    # Test LHS for sanity too (need to uncomment)
    grid_points = cell.gen_uniform_grids(mf.with_df.mesh)
    from pyscf.pbc.lib.kpts_helper import conj_mapping

    minus_k_map = conj_mapping(cell, kpts)
    iq, ik, ik_prime = np.random.randint(0, num_kpts, 3)
    # Get -Q index
    minus_iq = minus_k_map[iq]
    ik_minus_q = momentum_map[iq, ik]
    iGpq = G_unique[iq, ik]
    iGsr = G_unique[iq, ik_prime]
    ik_prime_minus_q = momentum_map[iq, ik_prime]
    # Sanity check G mappings
    assert np.allclose(kpts[ik] - kpts[ik_minus_q] - kpts[iq],
                       delta_Gs[iq][iGpq])
    assert np.allclose(
        kpts[ik_prime] - kpts[ik_prime_minus_q] - kpts[iq],
        delta_Gs[iq][iGsr],
    )
    # (pk qk-Q | rk'-Q sk') = (q k-Q p k | sk' rk'-Q)*
    ik_prime_minus_q = momentum_map[iq, ik_prime]
    # uncomment to check normal eris
    # kpt_pqrs = [ik, ik_minus_q, ik_prime_minus_q, ik_prime]
    # eri_pqrs = build_eri(mf, kpt_pqrs)
    # kpt_pqrs = [ik, ik_minus_q, ik_prime_minus_q, ik_prime]
    # kpt_pqrs = [ik_minus_q, ik, ik_prime, ik_prime_minus_q]
    # eri_qpsr = build_eri(mf, kpt_pqrs).transpose((1, 0, 3, 2))
    # Sanity check relationship
    # assert np.allclose(eri_pqrs, eri_qpsr.conj())
    # Now check how to index into correct G when Q is conjugated
    # We want to find (-Q) + G_pq_comp + (Q + Gpq) = 0,
    # Q + Gpq = kp - kq = q
    # so G_pq_comp = -((-Q) + (Q+Gpq))
    iGpq_comp = minus_Q_G_map_unique[minus_iq, ik]
    iGsr_comp = minus_Q_G_map_unique[minus_iq, ik_prime]
    # Check zeta symmetry: expect zeta[Q,G1,G2,m,n] =
    # zeta[-Q,G1_comp,G2_comp,m, n].conj()
    # Build refernce point zeta[Q,G1,G2,m,n]
    zeta_ref = kpt_thc.zeta[iq][iGpq, iGsr]
    zeta_test = kpt_thc.zeta[minus_iq][iGpq_comp, iGsr_comp]
    # F31 (pk qk-Q | rk'-Q sk') = (rk'-Q s k'| pk qk-Q)
    assert np.allclose(zeta_ref, zeta_test.conj())
    # Sanity check do literal minus signs (should be complex
    # conjugate)
    zeta_test = build_kpoint_zeta(
        mf.with_df,
        -kpts[iq],
        -delta_Gs[iq][iGpq],
        -delta_Gs[iq][iGsr],
        grid_points,
        kpt_thc.xi,
    )
    assert np.allclose(zeta_ref, zeta_test.conj())
    # (pk qk-Q | rk'-Q sk') = (rk'-Q s k'| pk qk-Q)
    # uncomment to check normal eris
    # kpt_pqrs = [ik_prime_minus_q, ik_prime, ik, ik_minus_q]
    # eri_rspq = build_eri(mf, kpt_pqrs).transpose((2, 3, 0, 1))
    # assert np.allclose(eri_pqrs, eri_rspq)
    # Check zeta symmetry: expect zeta[Q,G1,G2,m,n] =
    # zeta[-Q,G2_comp,G1_comp,m, n]
    zeta_test = kpt_thc.zeta[minus_iq][iGsr_comp, iGpq_comp]
    assert np.allclose(zeta_ref, zeta_test.T)
    # (pk qk-Q | rk'-Q sk') = (sk' r k'-Q| qk-Q pk)
    # uncomment to check normal eris
    # kpt_pqrs = [ik_prime, ik_prime_minus_q, ik_minus_q, ik]
    # eri_srqp = build_eri(mf, kpt_pqrs).transpose((3, 2, 1, 0))
    # assert np.allclose(eri_pqrs, eri_srqp.conj())
    # Check zeta symmetry: expect zeta[Q,G1,G2,m,n]
    # = zeta[Q,G2,G1,n, m].conj()
    zeta_test = kpt_thc.zeta[iq][iGsr, iGpq]
    assert np.allclose(zeta_ref, zeta_test.conj().T)


@pytest.mark.skipif(not HAVE_DEPS_FOR_RESOURCE_ESTIMATES,
                    reason='pyscf and/or jax not installed.')
def test_symmetry_of_G_maps():
    cell = gto.Cell()
    cell.atom = """
    C 0.000000000000   0.000000000000   0.000000000000
    C 1.685068664391   1.685068664391   1.685068664391
    """
    cell.basis = "gth-szv"
    cell.pseudo = "gth-hf-rev"
    cell.a = """
    0.000000000, 3.370137329, 3.370137329
    3.370137329, 0.000000000, 3.370137329
    3.370137329, 3.370137329, 0.000000000"""
    cell.unit = "B"
    cell.verbose = 0
    cell.build(parse_arg=False)

    kmesh = [3, 3, 3]
    kpts = cell.make_kpts(kmesh)
    momentum_map = build_momentum_transfer_mapping(cell, kpts)
    (
        G_vecs,
        G_map,
        _,
        delta_Gs,
    ) = build_g_vector_mappings_double_translation(cell, kpts, momentum_map)
    G_dict, _ = build_g_vectors(cell)
    num_kpts = len(kpts)
    lattice_vectors = cell.lattice_vectors()
    from pyscf.pbc.lib.kpts_helper import conj_mapping

    minus_k_map = conj_mapping(cell, kpts)
    # k1 - k2 = Q + G
    for iq in range(1, num_kpts):
        minus_iq = minus_k_map[iq]
        for ik in range(num_kpts):
            Gpq = G_vecs[G_map[iq, ik]]
            Gpq_comp = -(kpts[minus_iq] + kpts[iq] + Gpq)
            iGpq_comp = G_dict[tuple(get_miller(lattice_vectors, Gpq_comp))]
            G_indx_unique = [
                G_dict[tuple(get_miller(lattice_vectors, G))]
                for G in delta_Gs[minus_iq]
            ]
            if iq == 1:
                pass
            assert iGpq_comp in G_indx_unique
            for _ in range(num_kpts):
                # Check complement(miller_Gpq) = miller_Gpq_comp
                # Get indx of "complement" G in original set of 27
                iGsr_comp = G_dict[tuple(get_miller(lattice_vectors, Gpq_comp))]
                # Get index of unique Gs in original set of 27
                # Check complement is in set corresponding to zeta[-Q]
                assert iGsr_comp in G_indx_unique

    # Check minus Q mapping
    minus_Q_G_map, minus_Q_G_map_unique = build_minus_q_g_mapping(
        cell, kpts, momentum_map)
    for iq in range(1, num_kpts):
        minus_iq = minus_k_map[iq]
        for ik in range(num_kpts):
            Gpq = G_vecs[G_map[iq, ik]]
            Gpq_comp = -(kpts[minus_iq] + kpts[iq] + Gpq)
            iGpq_comp = G_dict[tuple(get_miller(lattice_vectors, Gpq_comp))]
            assert iGpq_comp == minus_Q_G_map[minus_iq, ik]
            indx_in_unique_set = minus_Q_G_map_unique[minus_iq, ik]
            Gpq_comp_from_map = delta_Gs[iq][indx_in_unique_set]
            assert np.allclose(Gpq_comp, Gpq_comp_from_map)


@pytest.mark.skipif(not HAVE_DEPS_FOR_RESOURCE_ESTIMATES,
                    reason='pyscf and/or jax not installed.')
def test_isdf_qrcp():
    cell = gto.Cell()
    cell.atom = """
    C 0.000000000000   0.000000000000   0.000000000000
    C 1.685068664391   1.685068664391   1.685068664391
    """
    cell.basis = "gth-szv"
    cell.pseudo = "gth-hf-rev"
    cell.a = """
    0.000000000, 3.370137329, 3.370137329
    3.370137329, 0.000000000, 3.370137329
    3.370137329, 3.370137329, 0.000000000"""
    cell.mesh = [11] * 3
    cell.unit = "B"
    cell.verbose = 0
    cell.build()

    kmesh = [1, 2, 1]
    kpts = cell.make_kpts(kmesh)
    mf = scf.KRHF(cell, kpts)
    mf.kernel()

    num_mo = mf.mo_coeff[0].shape[-1]
    num_thc = np.prod(cell.mesh)
    kpt_thc = solve_qrcp_isdf(mf, num_thc, single_translation=False)
    num_kpts = len(kpts)
    momentum_map = build_momentum_transfer_mapping(cell, kpts)
    for iq in range(1, num_kpts):
        for ikp in range(num_kpts):
            ikq = momentum_map[iq, ikp]
            for iks in range(num_kpts):
                ikr = momentum_map[iq, iks]
                kpt_pqrs = [kpts[ikp], kpts[ikq], kpts[ikr], kpts[iks]]
                mos_pqrs = [
                    mf.mo_coeff[ikp],
                    mf.mo_coeff[ikq],
                    mf.mo_coeff[ikr],
                    mf.mo_coeff[iks],
                ]
                eri_pqrs = mf.with_df.ao2mo(mos_pqrs, kpt_pqrs,
                                            compact=False).reshape(
                                                (num_mo,) * 4)
                eri_pqrs_isdf = build_eri_isdf_double_translation(
                    kpt_thc.chi,
                    kpt_thc.zeta,
                    iq,
                    [ikp, ikq, ikr, iks],
                    kpt_thc.g_mapping,
                )
                assert np.allclose(eri_pqrs, eri_pqrs_isdf)



# coverage: ignore
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""
Estimate the physical, logical, and Toffoli gate requirements for
single-factorization qubitization.  Single-Factorization uses
an LCU formed from a symmeterized Cholesky decomposition of the integrals
"""
from typing import Tuple

import numpy as np
from numpy.lib.scimath import arccos, arcsin
from sympy import factorint

from openfermion.resource_estimates.utils import QI
from openfermion.resource_estimates.utils import QR2 as QR2_of
from openfermion.resource_estimates.pbc.resources import (
    ResourceEstimates,
    QR3,
    QR2,
    QI2,
)


def compute_cost(
        num_spin_orbs: int,
        lambda_tot: float,
        num_aux: int,
        kmesh: list[int],
        dE_for_qpe: float = 0.0016,
        chi: int = 10,
) -> ResourceEstimates:
    """Determine fault-tolerant costs using single factorizated Hamiltonian.

    Light wrapper around _compute_cost to automate choice of stps paramter.

    Arguments:
        num_spin_orbs: the number of spin-orbitals
        lambda_tot: the lambda-value for the Hamiltonian
        num_sym_unique: number of symmetry unique terms kept in the sparse
            Hamiltonian
        dE_for_qpe: allowable error in phase estimation
        chi: the number of bits for the representation of the coefficients
    Returns:
        resources: sparse resources
    """
    # run once to determine stps parameter
    init_cost = _compute_cost(
        num_spin_orbs,
        lambda_tot,
        num_aux,
        dE_for_qpe,
        chi,
        20_000,
        kmesh[0],
        kmesh[1],
        kmesh[2],
    )
    steps = init_cost[0]
    final_cost = _compute_cost(
        num_spin_orbs,
        lambda_tot,
        num_aux,
        dE_for_qpe,
        chi,
        steps,
        kmesh[0],
        kmesh[1],
        kmesh[2],
    )
    estimates = ResourceEstimates(
        toffolis_per_step=final_cost[0],
        total_toffolis=final_cost[1],
        logical_qubits=final_cost[2],
    )
    return estimates


def _compute_cost(
        n: int,
        lam: float,
        M: int,
        dE: float,
        chi: int,
        stps: int,
        Nkx: int,
        Nky: int,
        Nkz: int,
        verbose: bool = False,
) -> Tuple[int, int, int]:
    """Determine fault-tolerant costs using SF decomposition in quantum chem

    Args:
        n: - the number of spin-orbitals. When using this for
                  k-point sampling, this is equivalent to N times N_k.
        lam: the lambda-value for the Hamiltonian
        M: The combined number of values of n
        dE:  allowable error in phase estimation
        chi: equivalent to aleph_1 and aleph_2 in the document, the
            number of bits for the representation of the coefficients
        stps: an approximate number of steps to choose the precision of
            single qubit rotations in preparation of the equal superposn state
        Nkx: num k-points in x-direction
        Nky: num k-points in y-direction
        Nkz: num k-points in z-direction
        verbose: do additional printing of intermediates?

    Returns:
        step_cost: Toffolis per step
        total_cost: Total number of Toffolis
        total_qubit_count: Total qubit count
    """
    nNk = (max(np.ceil(np.log2(Nkx)), 1) + max(np.ceil(np.log2(Nky)), 1) +
           max(np.ceil(np.log2(Nkz)), 1))
    Nk = Nkx * Nky * Nkz
    L = Nk * n**2 // 2

    # Number of trailing zeros by computing smallest prime factor exponent
    # Number of trailing zeros
    factors = factorint(L)
    eta = factors[min(list(sorted(factors.keys())))]
    if L % 2 == 1:
        eta = 0

    # Number of qubits for the first register
    nL = np.ceil(np.log2(L))

    # Number of qubits for p and q registers
    nN = np.ceil(np.log2(n // 2))

    nMN = np.ceil(np.log2(M * Nk + 1))

    oh = [0] * 20
    for p in range(20):
        # JJG note: arccos arg may be > 1
        # v = Round[2^p/(2*\[Pi])*ArcCos[2^nL/Sqrt[(L + 1)/2^\[Eta]]/2]];
        v = np.round(
            np.power(2, p + 1) / (2 * np.pi) *
            arccos(np.power(2, nMN) / np.sqrt(M * Nk + 1) / 2))
        #   oh[[p]] = stps*(1/N[Sin[3*ArcSin[Cos[v*2*\[Pi]/2^p]*Sqrt[(L +
        #   1)/2^\[Eta]]/2^nL]]^2] - 1) + 4*p];
        #   stps*(1/N[Sin[3*ArcSin[Cos[v*2*\[Pi]/2^p]*Sqrt[M*Nk + 1]/2^nMN]]^2]
        #   - 1) + 4*p]
        oh[p] = np.real(stps * (1 / (np.sin(3 * arcsin(
            np.cos(v * 2 * np.pi / np.power(2, p + 1)) * np.sqrt(M * Nk + 1) /
            np.power(2, nMN)))**2) - 1) + 4 * (p + 1))

    # Bits of precision for rotation
    br1 = int(np.argmin(oh) + 1)

    oh2 = [0] * 20
    for p in range(20):
        # JJG note: arccos arg may be > 1
        # v = Round[2^p/(2*\[Pi])*ArcCos[2^nL/Sqrt[(L + 1)/2^\[Eta]]/2]];
        v = np.round(
            np.power(2, p + 1) / (2 * np.pi) *
            arccos(np.power(2, nL) / np.sqrt(L / np.power(2, eta)) / 2))
        #   oh[[p]] = stps*(1/N[Sin[3*ArcSin[Cos[v*2*\[Pi]/2^p]*Sqrt[(L +
        #   1)/2^\[Eta]]/2^nL]]^2] - 1) + 4*p]; oh2[[p]] =
        #   stps*(1/N[Sin[3*ArcSin[Cos[v*2*\[Pi]/2^p]*Sqrt[L/2^\[Eta]]/2^nL]]^2]
        #   - 1) + 4*p];
        oh2[p] = np.real(stps * (1 / (np.sin(3 * arcsin(
            np.cos(v * 2 * np.pi / np.power(2, p + 1)) *
            np.sqrt(L / np.power(2, eta)) / np.power(2, nL)))**2) - 1) + 4 *
                         (p + 1))
    br2 = int(np.argmin(oh2) + 1)

    # Cost of preparing the equal superposition state on the 1st register in 1a
    cost1a = 2 * (3 * nMN + 2 * br1 - 9)

    # We have added two qubits for ell <> 0.
    # bMN = 2*(np.ceil(np.log2(Nk)) + np.ceil(np.log2(M))) + chi + 2
    bMN = nMN + 2 * np.ceil(np.log2(Nk)) + chi + 2

    # QROM costs for first register preparation in step 1(b)
    cost1b = QR3(M * Nk + 1, bMN)[-1] + QI(M * Nk + 1)[-1]

    # The inequality test of cost chi in step 1(c) and the controlled swap with
    # cost nL + 1 in step 1(d) (and their inversions).
    cost1cd = 2 * (chi + np.ceil(np.log2((Nk)) + np.ceil(np.log2(M)) + 1))

    # Cost of preparing equal superposition of p and q registers in step 2(a).
    cost2a = 4 * (3 * nL - 3 * eta + 2 * br2 - 9)

    # The output size for the QROM for the second state preparation.
    bp = 6 * np.ceil(np.log2(Nk) / 3) + 4 * nN + chi + 3

    # Cost of QROMs for state preparation on p and q in step 2 (c).
    cost2b = (QR2(M * Nk + 1, L, bp) + QI2(M * Nk + 1, L) + QR2(M * Nk, L, bp) +
              QI2(M * Nk, L))

    # The cost of the inequality test and controlled swap for the
    # quantum alias sampling in steps 2(c) and (d).
    cost2cd = 4 * (chi + 3 * np.ceil(np.log2(Nk) / 3) + 2 * nN + 1)

    # Computing k - Q and k' - Q, then below we perform adjustments for non -
    # modular arithmetic.
    cost3 = 4 * nNk
    if Nkx == 2**np.ceil(np.log2(Nkx)):
        cost3 = cost3 - 2 * np.ceil(np.log2(Nkx))
    if Nky == 2**np.ceil(np.log2(Nky)):
        cost3 = cost3 - 2 * np.ceil(np.log2(Nky))
    if Nkz == 2**np.ceil(np.log2(Nkz)):
        cost3 = cost3 - 2 * np.ceil(np.log2(Nkz))

    # The SELECT operation in step 4, which needs to be done twice.
    cost4 = 2 * (3 * n * Nk - 4)

    # The reflection used on the 2nd register in the middle, with cost in step 6
    cost6 = nL + chi + 5

    # Step 7 involves performing steps 2 to 5 again, which are accounted for
    # above, but there is one extra Toffoli for checking that l <> 0.
    cost7 = 4

    # The total reflection for the quantum walk in step 9.
    # with an extra + 1 for k - point sampling.
    cost9 = nMN + nL + 2 * chi + 4

    # The two Toffolis for the control for phase estimation and iteration,
    # given in step 10.
    cost10 = 2

    # The number of steps.
    iters = np.ceil(np.pi * lam / (dE * 2))

    # The total Toffoli costs for a step.
    cost = (cost1a + cost1b + cost1cd + cost2a + cost2b + cost2cd + cost3 +
            cost4 + cost6 + cost7 + cost9 + cost10)

    # Control for phase estimation and its iteration
    ac1 = 2 * np.ceil(np.log2(iters)) - 1

    # System qubits
    ac2 = n * Nk

    # First ell register that rotated and the flag for success
    ac3 = nMN + 2

    # State preparation on the first register
    ac4 = bMN + chi + 1

    # For p and q registers, the rotated qubit and the success flag
    ac5 = nL + 2

    kp = QR2_of(M * Nk + 1, L, bp)[:2]

    ac6 = chi

    # The equal superposition state for the second state preparation
    ac7 = np.max([br1, br2])

    # The phase gradient state
    ac8 = 4

    # The QROM on the p & q registers
    ac9 = (kp[0] * kp[1] * bp + np.ceil(np.log2(
        (M * Nk + 1) / kp[0])) + np.ceil(np.log2(L / kp[1])))

    if verbose:
        print("[*] Top of routine")
        print("  [+] eta = ", eta)
        print("  [+] nL = ", nL)
        print("  [+] nN = ", nN)

    total_qubit_count = ac1 + ac2 + ac3 + ac4 + ac5 + ac6 + ac7 + ac8 + ac9

    # Sanity checks before returning as int
    assert cost.is_integer()
    assert iters.is_integer()
    assert total_qubit_count.is_integer()

    step_cost = int(cost)
    total_cost = int(cost * iters)
    total_qubit_count = int(total_qubit_count)

    return step_cost, total_cost, total_qubit_count
